//=============================================================================
// RPG Maker MZ - Advanced Enemies
// ----------------------------------------------------------------------------
// (C)2025 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2025/02/15
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Advanced Enemies
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help advanced_enemies.js
 *
 * Advanced Enemies
 * 
 * Added tags:
 * <conditional_drop_rates: [SCRIPT_1, ..., SCRIPT_N]>
 * - Overwrites the drop rate of drop i by the eval result of SCRIPT i.
 * - When drop rate > 1, n objects will be dropped.
 * <conditional_gold: SCRIPT>
 * - Overwrites the gold drop with the eval result of SCRIPT.
 * Available variables: defaultGold (original gold drop)
 * <conditional_actions: return [CONDITION_1, ..., CONDITION_N]>
 * - Disables action i if CONDITION i evaluates to false.
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.AdvancedEnemies = {};

(($) => {
	// Inject logic to conditionally change item drop rates
	$.Game_Enemy_makeDropItems = Game_Enemy.prototype.makeDropItems;
	Game_Enemy.prototype.makeDropItems = function() {
		const enemy = this.enemy();
		if (enemy.meta.conditional_drop_rates) {
			if (!enemy.conditionalDropRates) {
				enemy.conditionalDropRates = new Function(enemy.meta.conditional_drop_rates);
			}

			const conditionalDropRates = enemy.conditionalDropRates.call(this);
			const rate = this.dropItemRate();
			return this.enemy().dropItems.reduce((r, di, index) => {
				const countDrops = Math.ceil(conditionalDropRates[index] * rate - Math.random());
				if (di.kind > 0 && countDrops > 0) {
					const drop = this.itemObject(di.kind, di.dataId);
					const drops = Array.from(Array(countDrops), () => drop)
					return r.concat(drops);
				} else {
					return r;
				}
			}, []);
		} else {
			return $.Game_Enemy_makeDropItems.call(this);
		}
	};

	// Inject logic to overwrite the default gold value
	$.Game_Enemy_gold = Game_Enemy.prototype.gold;
	Game_Enemy.prototype.gold = function() {
		const enemy = this.enemy();
		const defaultGold = $.Game_Enemy_gold.call(this);
		if (enemy.meta.conditional_gold) {
			if (!enemy.conditionalGold) {
				enemy.conditionalGold = new Function("defaultGold", enemy.meta.conditional_gold);
			}
			return enemy.conditionalGold.call(this, defaultGold);
		} else {
			return defaultGold;
		}
	};

	// Inject logic to check if an action can be used on a per-enemy basis
	$.Game_Enemy_isActionValid = Game_Enemy.prototype.isActionValid;
	Game_Enemy.prototype.isActionValid = function(action) {
		const enemy = this.enemy();
		if (enemy.meta.conditional_actions) {
			if (!enemy.conditionalActions) {
				enemy.conditionalActions = new Function(enemy.meta.conditional_actions);
			}

			const conditions = enemy.conditionalActions.call(this);
			const actionIndex = this.enemy().actions.indexOf(action);
			if (!conditions[actionIndex]) {
				return false;
			}
		}

		return $.Game_Enemy_isActionValid.call(this, action);
	};
})(AuraMZ.AdvancedEnemies);

