//=============================================================================
// RPG Maker MZ - PLUGIN NAME
// ----------------------------------------------------------------------------
// (C)2025 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2025/02/13
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Advanced Pictures
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help advanced_pictures.js
 *
 * Adds utility commands to simplify scenes with many pictures 
 * and transitions between them.
 * 
 * @param transitionTypes
 * @type struct<transitionType>[]
 * @desc A list of predefined picture transition parameters.
 * @default ["{\"name\":\"default\",\"duration\":\"60\",\"easingType\":\"0\"}","{\"name\":\"fast\",\"duration\":\"30\",\"easingType\":\"0\"}","{\"name\":\"instant\",\"duration\":\"1\",\"easingType\":\"0\"}"]
 *
 * @command attachChildren
 * @text Attach Children
 * @desc Attaches a list of pictures as a child of another.
 *
 * @arg parentId
 * @type number
 * @text ParentID
 * @desc ID of the parent picture
 * 
 * @arg childrenIds
 * @type number[]
 * @text ChildrenIDs
 * @desc IDs of the child pictures
 * 
 * @command transition
 * @text Transiton
 * @desc Fades one image out and the other image in.
 *
 * @arg oldPictureId
 * @type number
 * @text Old PictureID
 * @desc ID of the picture to fade out
 * 
 * @arg newPictureId
 * @type number
 * @text New PictureID
 * @desc ID of the picture to fade in
 * 
 * @arg duration
 * @type number
 * @text Duration
 * @desc Duration in frames
 * @default 30
 *
 * @arg easingType
 * @type combo
 * @option Constant
 * @value 0
 * @option Slow Start
 * @value 1
 * @option Slow End
 * @value 2
 * @option Slow Start and End
 * @value 3
 * @text Easing Type
 * @desc The easing type
 * @default 0
 * 
 * @arg waitForCompletion
 * @text Wait for Completion
 * @desc Whether the interpreter needs to wait for the transition to complete.
 * @type boolean
 * @default true
 * 
 * @command transitionTyped
 * @text Transiton (Typed)
 * @desc Fades one image out and the other image in using a predefined transition setting.
 *
 * @arg oldPictureId
 * @type number
 * @text Old PictureID
 * @desc ID of the picture to fade out
 * 
 * @arg newPictureId
 * @type number
 * @text New PictureID
 * @desc ID of the picture to fade in
 * 
 * @arg transitionType
 * @type text
 * @text Transition Type
 * @default default
 * @desc The name of the transition type
 * 
 * @arg waitForCompletion
 * @text Wait for Completion
 * @desc Whether the interpreter needs to wait for the transition to complete.
 * @type boolean
 * @default true
 * 
 * @command copyPictureHierarchy
 * @text Copy Picture Hierarchy
 * @desc Creates a copy of the target picture ID and all children (recursively).
 *
 * @arg copyIdSrc
 * @type number
 * @text Copy ID Source
 * @desc The ID of the source parent picture.
 * 
 * @arg copyIdDst
 * @type number
 * @text Copy ID Destination
 * @desc The ID of the new parent picture. All children IDs will be relative to the new destination ID.
 * 
 * @arg copyOpacity
 * @type number
 * @text Copy Opacity
 * @default 255
 * @desc The opacity of the copy.
 */

/*~struct~transitionType:
 *
 * @param name
 * @type text
 * @text Name
 * @desc Name of this transition type
 *
 * @param duration
 * @type number
 * @text Duration
 * @desc Duration in frames
 * @default 30
 *
 * @param easingType
 * @type combo
 * @option Constant
 * @value 0
 * @option Slow Start
 * @value 1
 * @option Slow End
 * @value 2
 * @option Slow Start and End
 * @value 3
 * @text Easing Type
 * @desc The easing type
 * @default 0
 *
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.AdvancedPictures = {};

(($) => {
	$.PLUGIN_NAME = "advanced_pictures";
	$.Params = PluginManager.parameters($.PLUGIN_NAME);
	$.transitionTypes = JSON.parse($.Params["transitionTypes"]).map(jsonTransitionType => JSON.parse(jsonTransitionType));

	// Registers a command to attach a picture as a child to another picture
	PluginManager.registerCommand($.PLUGIN_NAME, "attachChildren", args => {
		const parentId = parseInt(args.parentId);
		const childrenIds = JSON.parse(args.childrenIds);
		for (const childId of childrenIds) {
			const childPicture = $gameScreen.picture(parseInt(childId));
			childPicture._parent = parentId;
		}
	});

	// Command to transition between two pictures
	function commandTransition(args) {
		const oldPictureId = parseInt(args.oldPictureId);
		const newPictureId = parseInt(args.newPictureId);
		const oldPicture = $gameScreen.picture(oldPictureId);
		const newPicture = $gameScreen.picture(newPictureId);
		const duration = parseInt(args.duration);
		const easingType = parseInt(args.easingType);

		oldPicture._targetOpacity = 0;
		oldPicture._duration = duration;
		oldPicture._wholeDuration = duration;
		oldPicture._easingExponent = 2;
		oldPicture._easingType = easingType;
		newPicture._targetOpacity = 255;
		newPicture._duration = duration;
		newPicture._wholeDuration = duration;
		newPicture._easingExponent = 2;
		newPicture._easingType = easingType;

		const waitForCompletion = eval(args.waitForCompletion);
		if (waitForCompletion) {
			this.wait(duration);
		}
	}

	// Command to transition between two pictures using a predefined transition type
	function commandTransitionTyped(args) {
		const name = args.transitionType;
		for (const transitionType of $.transitionTypes) {
			if (transitionType.name == name) {
				args.duration = transitionType.duration;
				args.easingType = transitionType.easingType;
			}
		}
		commandTransition.call(this, args);
	}

	// Register commands
	PluginManager.registerCommand($.PLUGIN_NAME, "transition", commandTransition);
	PluginManager.registerCommand($.PLUGIN_NAME, "transitionTyped", commandTransitionTyped);


	// Registers a command to copy a hierarchy of pictures
	PluginManager.registerCommand($.PLUGIN_NAME, "copyPictureHierarchy", args => {
		const copyIdSrc = parseInt(args.copyIdSrc);
		const copyIdDst = parseInt(args.copyIdDst);
		const copyOpacity = parseInt(args.copyOpacity);
		$.copyPictureHierarchy(copyIdSrc, copyIdDst);
		$gameScreen.picture(copyIdDst)._opacity = copyOpacity;
	});

	// Performsa single copy operation and calls itself recurseviely
	$.copyPictureHierarchy = function(copyIdSrc, copyIdDst) {
		const picture = $gameScreen.picture(copyIdSrc);
		$gameScreen.showPicture(copyIdDst,
			picture._name, picture._origin, picture._x, picture._y, picture._scaleX, picture._scaleY, picture._opacity, picture._blendMode
		);

		const childPictureIds = picture.childrenIds();
		for (const childPictureId of childPictureIds) {
			const idDelta = childPictureId - copyIdSrc;
			const newCopyIdDest = copyIdDst + idDelta;
			$.copyPictureHierarchy(childPictureId, newCopyIdDest);
			const childPictureDest = $gameScreen.picture(newCopyIdDest);
			childPictureDest._parent = copyIdDst;
		}
	}

	// Getter for the parent picture ID
	Game_Picture.prototype.parent = function() {
		return $gameScreen.picture(this._parent);
	};

	// Gets the list children Ids of this picture
	Game_Picture.prototype.childrenIds = function() {
		const childrenIds = [];
		for (let i = 0; i < $gameScreen.maxPictures(); ++i) {
			const childPicture = $gameScreen.picture(i);
			if (childPicture?.parent() === this) {
				childrenIds.push(i);
			}
		}

		return childrenIds;
	};

	// Inject logic to inherit parent from previous image at the same picture ID
	// Important so changing bitmaps doesn't break the hierarchy
	$.Game_screen_showPicture = Game_Screen.prototype.showPicture
	Game_Screen.prototype.showPicture = function(
		pictureId, name, origin, x, y, scaleX, scaleY, opacity, blendMode
	) {
		const oldPicture = $gameScreen.picture(pictureId);
		$.Game_screen_showPicture.call(this, pictureId, name, origin, x, y, scaleX, scaleY, opacity, blendMode);
		if (oldPicture && oldPicture._parent) {
			const newPicture = $gameScreen.picture(pictureId);
			newPicture._parent = oldPicture._parent;
		}
	};

	// Inherit parent x position
	$.Game_Picture_x = Game_Picture.prototype.x;
	Game_Picture.prototype.x = function() {
		let x = $.Game_Picture_x.call(this);
		if (this._parent) {
			x += $gameScreen.picture(this._parent).x();
		}
		return x;
	};

	// Inherit parent y position
	$.Game_Picture_y = Game_Picture.prototype.y;
	Game_Picture.prototype.y = function() {
		let y = $.Game_Picture_y.call(this);
		if (this._parent) {
			y += $gameScreen.picture(this._parent).y();
		}
		return y;
	};

	// Inherit parent x scale
	$.Game_Picture_scaleX = Game_Picture.prototype.scaleX;
	Game_Picture.prototype.scaleX = function() {
		let scaleX = $.Game_Picture_scaleX.call(this);
		if (this._parent) {
			scaleX *= $gameScreen.picture(this._parent).scaleX() / 100.0;
		}
		return scaleX;
	};

	// Inherit parent y scale
	$.Game_Picture_scaleY = Game_Picture.prototype.scaleY;
	Game_Picture.prototype.scaleY = function() {
		let scaleY = $.Game_Picture_scaleY.call(this);
		if (this._parent) {
			scaleY *= $gameScreen.picture(this._parent).scaleY() / 100.0;
		}
		return scaleY;
	};

	// Inherit parent opacity
	$.Game_Picture_opacity = Game_Picture.prototype.opacity;
	Game_Picture.prototype.opacity = function() {
		let opacity = $.Game_Picture_opacity.call(this);
		if (this._parent) {
			opacity *= $gameScreen.picture(this._parent).opacity() / 255.0;
		}
		return opacity;
	};

	// Inherit parent tone
	$.Game_Picture_tone = Game_Picture.prototype.tone;
	Game_Picture.prototype.tone = function() {
		let tone = $.Game_Picture_tone.call(this);
		if (tone && this._parent) {
			let parentTone = $gameScreen.picture(this._parent).tone();
			if (parentTone) {
				for (let i = 0; i < tone.length; ++i) {
					tone[i] *= parentTone[i] / 255.0
				}
			}
		}
		return tone;
	};

	// Inherit parent angle
	$.Game_Picture_angle = Game_Picture.prototype.angle;
	Game_Picture.prototype.angle = function() {
		let angle = $.Game_Picture_angle.call(this);
		if (this._parent) {
			angle += $gameScreen.picture(this._parent).angle();
		}
		return angle;
	};

	// Inject logic to erase child pictures when erasing parent
	$.Game_Screen_erasePicture = Game_Screen.prototype.erasePicture;
	Game_Screen.prototype.erasePicture = function(pictureId) {
		const picture = this.picture(pictureId);
		if (picture) {
			const childrenIds = picture.childrenIds();
			childrenIds.forEach(childId => this.erasePicture(childId));
		}
		$.Game_Screen_erasePicture.call(this, pictureId);
	};
})(AuraMZ.AdvancedPictures);

