//=============================================================================
// RPG Maker MZ - Analytics
// ----------------------------------------------------------------------------
// (C)2021 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2025/02/11
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Analytics
 * @author Gaurav Munjal
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help analytics.js
 *
 * Tracking for analytics
 * 
 * Dependencies:
 * - ../../libs/GameAnalytics.js
 * - ../../env.js
 * 
 * After:
 * - newgame_plus_skills.js 
 * - TLB_SKACraftMenu.js
 * - quests.js
 * 
 * @requiredAssets js/env.js
 * @requiredAssets js/libs/GameAnalytics.js
 *
 * @param analyticsQuestion_cte
 * @desc Analytics Question
 * @type string
 * 
 * @param difficultyVar
 * @text Difficulty Variable
 * @desc ID of variable to use for the game's difficulty.
 * @default 9
 * 
 * @param dayVar
 * @text Day Variable
 * @desc ID of variable to use for the current day.
 * @default 2
 *
 * @param agreeText_cte
 * @text Agree Text
 * @desc Text to use for the 'Agree' choice option.
 * @default Agree
 *
 * @param declineText_cte
 * @text Decline Text
 * @desc Text to use for the 'Decline' choice option.
 * @default Decline
 *
 * @param analyticsText_cte
 * @text Analytics Text
 * @desc Text to use for the 'Analytics' option.
 * @default Analytics
 *
 * @command sendProgressionEvent
 * @text Send Progression Event
 * @desc Sends a progression event status, category:subcategory:difficulty,day.
 * E.g. quests fire the progressiong event Complete,Quest:QuestTitle:Difficulty,day
 *
 * @arg status
 * @type number
 * @text Status
 * @desc The status. 0: Start. 1: Complete. 2: Fail.
 *
 * @arg category
 * @type string
 * @text Category
 * @desc The progression event category. (E.g. Quest)
 *
 * @arg subcategory
 * @type string
 * @text Subcategory
 * @desc The progression event category. (E.g. QuestTitle)
 *
 * @command sendDesignEvent
 * @text Send Design Event
 * @desc Sends a design event. Event format is category1:category2:category3:category4 with the hierarchy being 5 deep max.
 * May optionally have a score.
 *
 * @arg categoryHierarchy
 * @type string[]
 * @min 1
 * @max 5
 * @text Category Hierarchy
 * @desc The design event hierarchy. Individual entries are eval'ed.
 * Max 5 entries.
 *
 * @arg score
 * @type string
 * @text Score
 * @desc Optional score. Eval'ed.
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.Analytics = {};

class Window_AcceptChoiceList extends Window_ChoiceList {
	windowY() {
		return super.windowY() + 50;
	}
}

// Scene that asks for permission to collect analytics information
class Scene_Analytics extends Scene_Message {
	create() {
		super.create();
		this.createWindowLayer();
		this.createAllWindows();
	}

	createChoiceListWindow() {
		this._choiceListWindow = new Window_AcceptChoiceList();
		this.addWindow(this._choiceListWindow);
	};

	// Inject showing a message box for collecting analytics
	start() {
		if (ConfigManager.telemetry || ConfigManager.initializedTelemetry) {
			this.gotoTitle();
			return;
		}

		super.start();
		$gameMessage.setBackground(1);
		$gameMessage.setPositionType(1);
		for (const line of AuraMZ.Analytics.analyticsquestion.split("\\n")) {
			$gameMessage.add(line);
		}

		const choices = [AuraMZ.Analytics.agree, AuraMZ.Analytics.decline];
		$gameMessage.setChoices(choices, 0, -1);
		$gameMessage.setChoiceBackground(1);
		$gameMessage.setChoicePositionType(1);
		this._choiceListWindow
		$gameMessage.setChoiceCallback(n => {
			ConfigManager.telemetry = n == 0;
			ConfigManager.initializedTelemetry = true;
			ConfigManager.save();
			this.gotoTitle();
		});
	}

	gotoTitle() {
		SceneManager.goto(Scene_Title);
	}
}

(($) => {
	const PLUGIN_NAME = "analytics";
	const PARAMS = PluginManager.parameters(PLUGIN_NAME);
	$.analytics_question = PARAMS["analyticsQuestion"];
	$.difficulty_var = parseInt(PARAMS["difficultyVar"]);
	$.day_var = parseInt(PARAMS["dayVar"]);

	Object.defineProperties(AuraMZ.Analytics, {
		"agree": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-agreeText_cte`) || PARAMS.agreeText_cte },
		"decline": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-declineText_cte`) || PARAMS.declineText_cte },
		"analyticscommand": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-analyticsText_cte`) || PARAMS.analyticsText_cte },
		"analyticsquestion": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-analyticsQuestion_cte`) || PARAMS.analyticsQuestion_cte }
	});

	function ensureConfigLoaded() {
		return new Promise((resolve) => {
			(function waitForConfigLoaded() {
				if (SceneManager._scene?.isReady() && ConfigManager.isLoaded() && $gameTemp) {
					return resolve();
				}
				setTimeout(waitForConfigLoaded, 50);
			})();
		});
	}

	// get a value for key from env file
	function getEnv(key) {
		return envKeys[key];
	}

	ensureConfigLoaded().then(() => {
		if (ConfigManager.telemetry && window.gameanalytics) {
			const ga = gameanalytics.GameAnalytics;
			ga.configureBuild(AuraMZ.gameVersion);
			if ($dataSystem) {
				ga.configureAvailableResourceCurrencies([$dataSystem.currencyUnit]);
			}
			ga.configureAvailableResourceItemTypes(["Shop"]);
			if ($gameTemp.isPlaytest()) {
				// Log events for debugging
				GameAnalytics("setEnabledInfoLog", true);

				// Get GAME_ANALYTICS_SECRET_KEY from your account at www.gameanalytics.com
				// This must be set in your environment while building the game.
				GameAnalytics("initialize", getEnv("GAME_ANALYTICS_API_KEY_TEST"), getEnv("GAME_ANALYTICS_SECRET_KEY_TEST"));
			} else {
				// Get GAME_ANALYTICS_SECRET_KEY from your account at www.gameanalytics.com
				// This must be set in your environment while building the game.
				GameAnalytics("initialize", getEnv("GAME_ANALYTICS_API_KEY_PROD"), getEnv("GAME_ANALYTICS_SECRET_KEY_PROD"));
			}

			$.addCustomEvents();
			if (AuraMZ.Deploy) {
				ga.addDesignEvent(
					`StartUp:${AuraMZ.Deploy.ci_commit_ref_name}:${AuraMZ.Deploy.ci_commit_sha}:${AuraMZ.Deploy.ci_job_id}`
				);
			}

		}
	});

	// Inject the analytics scene
	Scene_Disclaimer.prototype.gotoTitle = function() {
		SceneManager.goto(Scene_Analytics);
	}

	// add telemetry option to General Options
	const _Window_Options_addGeneralOptions = Window_Options.prototype.addGeneralOptions;
	Window_Options.prototype.addGeneralOptions = function() {
		_Window_Options_addGeneralOptions.call(this);
		this.addCommand($.analyticscommand, "telemetry");
	};

	Window_Options.prototype.isTelemetrySymbol = function(symbol) {
		return symbol === 'telemetry';
	};

	Window_Options.prototype.telemetrySymbol = function() {
		return 'telemetry';
	};

	const ConfigManager_makeData = ConfigManager.makeData;
	ConfigManager.makeData = function() {
		const config = ConfigManager_makeData.call(this);
		config.telemetry = this.telemetry;
		config.initializedTelemetry = this.initializedTelemetry;
		return config;
	};

	const ConfigManager_applyData = ConfigManager.applyData;
	ConfigManager.applyData = function(config) {
		ConfigManager_applyData.call(this, config);
		this.telemetry = this.readFlag(config, "telemetry", false);
		this.initializedTelemetry = this.readFlag(config, "initializedTelemetry", false);
	};

	$.addCustomEvents = function() {
		const ga = gameanalytics.GameAnalytics;

		// Registers a command to send a progression event
		PluginManager.registerCommand(PLUGIN_NAME, "sendProgressionEvent", args => {
			let status = parseInt(args.status);
			switch (status) {
				case 0:
					status = gameanalytics.EGAProgressionStatus.Start
					break;
				case 1:
					status = gameanalytics.EGAProgressionStatus.Complete;
					break;
				case 2:
					status = gameanalytics.EGAProgressionStatus.Fail;
					break;
				default:
					console.log("Unknown progression status!");
			}

			ga.addProgressionEvent(
				status,
				args.category,
				args.subcategory,
				$gameVariables.value($.difficulty_var), // difficulty
				$gameVariables.value($.day_var)  // day WILL NOT BE PROCESSED SERVER SIDE FOR NON-COMPLETION EVENTS!
			);
		});

		// Registers a command to send a design event
		PluginManager.registerCommand(PLUGIN_NAME, "sendDesignEvent", args => {
			const categoryHierarchy = JSON.parse(args.categoryHierarchy);
			const eventID = categoryHierarchy
				.map(category => eval(category))
				.reduce((res, category) => res + ":" + category);
			if (args.score && args.score != "") {
				ga.addDesignEvent(eventID, eval(args.score));
			} else {
				ga.addDesignEvent(eventID);
			}
		});

		// battle
		const _BattleManager_endBattle = BattleManager.endBattle;
		BattleManager.endBattle = function(result) {
			let value;
			if (result === 0) {
				value = "victory";
			} else if (this._escaped) {
				value = "escaped";
			} else {
				value = "defeat";
			}
			const troopId = $gameTroop._troopId;
			const difficulty = $gameVariables.value($.difficulty_var);

			ga.addDesignEvent(`Battle:${troopId}:${difficulty}:${value}`, result);

			_BattleManager_endBattle.call(this, result);
		};

		// quest
		if (AuraMZ.Quests) {
			const _QuestManager_onAddQuest = QuestManager.onAddQuest;
			QuestManager.onAddQuest = function(quest) {
				_QuestManager_onAddQuest.call(this, quest);

				const status = gameanalytics.EGAProgressionStatus.Start;
				ga.addProgressionEvent(
					status,
					"Quest",
					$gameQuests._data[quest._key], // quest ID
					$gameVariables.value($.difficulty_var), // difficulty
					$gameVariables.value($.day_var)  // day
				);
			};

			const _QuestManager_onSetQuestStatus = QuestManager.onSetQuestStatus;
			QuestManager.onSetQuestStatus = function(quest) {
				_QuestManager_onSetQuestStatus.call(this, quest);

				let status = gameanalytics.EGAProgressionStatus.Complete
				if (quest._status !== 1) {
					status = gameanalytics.EGAProgressionStatus.Fail
				}

				ga.addProgressionEvent(
					status,
					"Quest",
					$gameQuests._data[quest._key], // quest ID
					$gameVariables.value($.difficulty_var), // difficulty
					$gameVariables.value($.day_var)  // day
				);
			};
		}

		// Toggle an option
		const _Window_Options_processOk = Window_Options.prototype.processOk;
		Window_Options.prototype.processOk = function() {
			_Window_Options_processOk.call(this);
			const index = this.index();
			const symbol = this.commandSymbol(index);
			const value = this.getConfigValue(symbol);

			ga.addDesignEvent(`OptionToggled:${symbol}:${value}`, value | 0);
		};

		if (Imported.TLB_SKACraftMenu) {
			const _Scene_Craft_removeIngredients = Scene_Craft.prototype.removeIngredients;
			Scene_Craft.prototype.removeIngredients = function(item, number) {
				_Scene_Craft_removeIngredients.call(this, item, number);
				for (const ingredient of item.ingredients) {
					for (let i = 0; i < ingredient.quantity * number; ++i) {
						ga.addDesignEvent(
							`Item:Ingredient:${$dataItems[ingredient.itemId].name}:${$gameVariables.value($.difficulty_var)}`,
							ingredient.itemId
						);
					}
				}
			}
		}

		const _Scene_Shop_doBuy = Scene_Shop.prototype.doBuy;
		Scene_Shop.prototype.doBuy = function(number) {
			_Scene_Shop_doBuy.call(this, number);
			const totalPrice = number * this.buyingPrice();
			ga.addResourceEvent(gameanalytics.EGAResourceFlowType.Sink, $dataSystem.currencyUnit, totalPrice, "Shop", this._item.name);
		};

		const _Scene_Shop_doSell = Scene_Shop.prototype.doSell;
		Scene_Shop.prototype.doSell = function(number) {
			_Scene_Shop_doSell.call(this, number);
			for (let i = 0; i < number; ++i) {
				ga.addDesignEvent(`Item:Sell:${this._item.name}:${$gameVariables.value($.difficulty_var)}`, this._item.id);
			}
		};

		const _Game_Battler_consumeItem = Game_Battler.prototype.consumeItem;
		Game_Battler.prototype.consumeItem = function(item) {
			_Game_Battler_consumeItem.call(this, item);
			ga.addDesignEvent(`Item:Consume:${item.name}:${$gameVariables.value($.difficulty_var)}`, item.id);
		};

		// Turn on carried over skills
		if (AuraMZ.carryOverSkills) {
			const _Game_Actor_learnCarryOverSkills = Game_Actor.prototype.learnCarryOverSkills;
			Game_Actor.prototype.learnCarryOverSkills = function() {
				_Game_Actor_learnCarryOverSkills.call(this);
				for (const skillID of AuraMZ.selectedCarryOverSkills) {
					ga.addDesignEvent(
						`NewGamePlus:Skill:${skillID}:${$gameVariables.value($.difficulty_var)}`,
						parseInt($dataSkills[skillID].meta.score_cost)
					);
				}
			};
		}

		// Inject sending error incident
		const _Graphics_printError = Graphics.printError;
		Graphics.printError = function(name, message, error = null) {
			_Graphics_printError.call(this, name, message, error);
			if (!$gameTemp.isPlaytest()) {
				try {
					const errorMessage = name + ":" + error?.stack ? error.stack : message;
					const currentInterpreter = $gameMap?.getCurrentInterpreter();
					const gameState =
						"\n-----------------------\n"
						+ "Map ID: " + $gameMap?.mapId() + "\n"
						+ "Running Event ID: " + currentInterpreter?.eventId() + "\n"
						+ "Interpreter Index: " + currentInterpreter?._index + "\n";
					ga.addErrorEvent(gameanalytics.EGAErrorSeverity.Critical, errorMessage + gameState);
				} catch (e) {
					// Don't want to cause a new error if there is an error in the reporting
				}
			}
		};
	}
})(AuraMZ.Analytics);
