//=============================================================================
// RPG Maker MZ - Blocking Fog
// ----------------------------------------------------------------------------
// (C)2021 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2024/09/25
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Blocking Fog
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help blocking_fog.js
 *
 * This plugin coveres regions which do not fulfill an unlock condition with a fog graphic.
 * Primarily intended to be used on the world map.
 *
 * Added tags:
 * <blocking_fog:CONDITION>
 * - If CONDITION evaluates to true, applies blocking fog to the map with this notetag.
 *
 *
 * @param defaultEnabled
 * @text Default Enabled
 * @desc Whether the blocking fog option will be enabled by default
 * @type boolean
 * @default true
 *
 * @param regionUnblockConditions
 * @text Region Unblock Conditions
 * @desc List of conditions that remove the blocking fog for a chosen region.
 * @type struct<regionUnblockCondition>[]
 * 
 * @param optionText_cte
 * @text Option Text
 * @desc The text to use for the 'World Fog' option.
 * @default World Fog
 *
 */
/*~struct~regionUnblockCondition:
*
* @param regionID
* @text Region ID
* @type number
* @desc All tiles of the region ID would become unblocked.
*
* @param condition
* @text Condition
* @desc The condition under which the region becomes unblocked.
*
*/

window.AuraMZ = window.AuraMZ || {};
AuraMZ.BlockingFog = {};

(($) => {
	const PLUGIN_NAME = "blocking_fog";
	const PARAMS = PluginManager.parameters(PLUGIN_NAME);

	Object.defineProperties($, {
		"worldfog": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-optionText_cte`) || PARAMS.optionText_cte }
	});

	$.defaultEnabled = PARAMS["defaultEnabled"] === 'true';
	$.regionUnblockConditions = [];
	
	const paramRegionUnblockConditions = PARAMS["regionUnblockConditions"] !== "" ? JSON.parse(PARAMS["regionUnblockConditions"]) : [];
	for (const jsonRegionUnblockCondition of paramRegionUnblockConditions) {
		const regionUnblockCondition = JSON.parse(jsonRegionUnblockCondition);
		$.regionUnblockConditions[regionUnblockCondition.regionID] = regionUnblockCondition.condition;
	}

	ConfigManager.blockingFog = $.defaultEnabled;

	// Check if blocking fog is enabled on this map
	Game_Map.prototype.isBlockingFogEnabled = function() {
		return ConfigManager.blockingFog && $dataMap.meta?.blocking_fog && eval($dataMap.meta.blocking_fog)
	};

	// Check if a region is unlocked
	Game_Map.prototype.isRegionUnlocked = function(regionId) {
		return regionId != 0 && eval($.regionUnblockConditions[regionId]);
	};

	const _Spriteset_Map_createLowerLayer = Spriteset_Map.prototype.createLowerLayer;
	Spriteset_Map.prototype.createLowerLayer = function() {
		_Spriteset_Map_createLowerLayer.call(this);
		if ($gameMap.isBlockingFogEnabled()) {
			this.createBlockingFogMask();
		}
	};

	// Creates the fog mask so that unlocked regions will not receive the fog effect
	Spriteset_Map.prototype.createBlockingFogMask = function() {
		const tileWidth = $gameMap.tileWidth();
		const tileHeight = $gameMap.tileHeight();
		const mapWidth = $gameMap.width();
		const mapHeight = $gameMap.height();
		const fogMask = new Bitmap(mapWidth * tileWidth, mapHeight * tileHeight);

		for (let x = 0; x < mapWidth; x++) {
			for (let y = 0; y < mapHeight; y++) {
				const regionId = $gameMap.regionId(x, y);
				if (!$gameMap.isRegionUnlocked(regionId)) {
					fogMask.fillRect(x * tileWidth, y * tileHeight, tileWidth, tileHeight, 'black');
				}
			}
		}

		this._blockingFogSpriteMask = new Sprite();
		this._blockingFogSpriteMask.bitmap = fogMask;
		const blurFilter = new PIXI.filters.BlurFilter();
		this._blockingFogSpriteMask.filters = [blurFilter];
		this._blockingFogSpriteMask.opacity = 240;
		this._tilemap.addChild(this._blockingFogSpriteMask);
		$.sprite = this._blockingFogSpriteMask;
	};

	const _Spriteset_Map_update = Spriteset_Map.prototype.update;
	Spriteset_Map.prototype.update = function() {
		_Spriteset_Map_update.call(this);
		this.updateBlockingFog();
	};

	Spriteset_Map.prototype.updateBlockingFog = function() {
		if (this._blockingFogSpriteMask) {
			this._blockingFogSpriteMask.x = -$gameMap.displayX() * $gameMap.tileWidth();
			this._blockingFogSpriteMask.y = -$gameMap.displayY() * $gameMap.tileHeight();
			this.animateFog();
		}
	};

	// Apply opacity and position fluctation for animation effect
	Spriteset_Map.prototype.animateFog = function() {
		const fogOpacitySpeed = 0.1;
		this._blockingFogSpriteMask.filters[0].blur = 8 + 40 * Math.sin(Date.now() / 1000) * fogOpacitySpeed;
	};

	// Inject logic to make fogged regions impassable
	const _Game_CharacterBase_canPass = Game_CharacterBase.prototype.canPass;
	Game_CharacterBase.prototype.canPass = function(x, y, d) {
		if ($gameMap.isBlockingFogEnabled()) {
			const x2 = $gameMap.roundXWithDirection(x, d);
			const y2 = $gameMap.roundYWithDirection(y, d);
			const regionId = $gameMap.regionId(x2, y2);
			if (!$gameMap.isRegionUnlocked(regionId)) {
				return false;
			}
		}

		return _Game_CharacterBase_canPass.call(this, x, y, d);
	};

	// Inject saving the blocking fog setting
	const ConfigManager_makeData = ConfigManager.makeData;
	ConfigManager.makeData = function() {
		const config = ConfigManager_makeData.call(this);
		config.blockingFog = this.blockingFog;
		return config;
	};

	// Injects the logic to update the blocking fog seetinga
	const _ConfigManager_applyData = ConfigManager.applyData;
	ConfigManager.applyData = function(config) {
		_ConfigManager_applyData.call(this, config);
		this.blockingFog = this.readFlag(config, "blockingFog", $.defaultEnabled);
	};

	// Insert the blocking fog option
	const _Window_Options_addGeneralOptions = Window_Options.prototype.addGeneralOptions;
	Window_Options.prototype.addGeneralOptions = function() {
		_Window_Options_addGeneralOptions.call(this);
		this.addCommand($.worldfog, "blockingFog", true);
	};

	// Increase number of option commands
	const _Scene_Options_maxCommands = Scene_Options.prototype.maxCommands;
	Scene_Options.prototype.maxCommands = function() {
		return _Scene_Options_maxCommands.call(this) + 1;
	};

	const Window_Options_processOk = Window_Options.prototype.processOk;
	Window_Options.prototype.processOk = function() {
		const index = this.index();
		const symbol = this.commandSymbol(index);
		if (symbol === "blockingFog") {
			const oldFlag = this.getConfigValue(symbol);
			this.changeValue(symbol, !oldFlag);
		} else {
			Window_Options_processOk.call(this);
		}
	};
})(AuraMZ.BlockingFog);