//=============================================================================
// RPG Maker MZ - Credits
// ----------------------------------------------------------------------------
// (C)2021 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2024/10/11
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Credits
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help credits.js
 *
 * This plugin offers functionality for displaying credits information.
 * The credits file is expected to be in a JSON format. The expected structure is:
 * [
 * 	{
 * 		"order" : 0,
 * 		"sectionName" : "SECTION_NAME"
 * 		"contributions" : [
 * 			"CONTRIBUTOR" : {
 * 				"contribution" : CONTRIBUTION,
 * 				"author" : AUTHOR,
 * 				"contribution" : CONTRIBUTION,
 * 			}
 * 		]
 * 	}
 * ]
 *
 * Dependencies:
 * - event_utils.js
 *
 * @param creditsFile
 * @text Credits File
 * @type text
 * @desc JSON credits file containing credits information
 * @default CREDITS.json
 *
 * @param creditsScrollSpeed
 * @text Credits Scroll Speed
 * @type number
 * @desc The scroll speed of the credits
 * @default 2
 *
 * @command playCredits
 * @text Play Credits
 * @desc Plays the credits from the provided credits file
 *
 */

window.AuraMZ = window.AuraMZ || {};

(() => {
	const PLUGIN_ID = "credits";
	const PARAMS = PluginManager.parameters(PLUGIN_ID);
	const CREDITS_FILE = PARAMS["creditsFile"];
	const CREDITS_SCROLL_SPEED = parseInt(PARAMS["creditsScrollSpeed"]);

	// Overwrite is busy check so that the interpreter will wait
	// for the credits to load
	const _Game_Message_isBusy = Game_Message.prototype.isBusy;
	Game_Message.prototype.isBusy = function() {
		return _Game_Message_isBusy.call(this) || this._isLoadingCredits;
	}

	DataManager.loadCredits = function() {
		return new Promise((resolve, reject) => {
			const xhr = new XMLHttpRequest();
			const name = "dataCredits";
			xhr.open("GET", CREDITS_FILE);
			xhr.overrideMimeType("application/json");
			xhr.onload = () => (xhr.status < 400) ? resolve(JSON.parse(xhr.responseText)) : reject(this.onXhrError(name, src, url));
			xhr.onerror = () => reject(this.onXhrError(name, CREDITS_FILE, CREDITS_FILE));
			xhr.send();
		});
	};

	// Plays the credits
	PluginManager.registerCommand(PLUGIN_ID, "playCredits", _ => {
		if ($gameMessage.isBusy()) {
			return false;
		}

		$gameMessage._isLoadingCredits = true;
		const interpreter = $gameMap.getCurrentInterpreter();
		interpreter.setWaitMode("message");

		DataManager.loadCredits().then(credits => {
			$gameMessage.setScroll(CREDITS_SCROLL_SPEED, false);
			for (const section of credits) {
				$gameMessage.add("\\{" + section.sectionName + "\\}");
				const contributors = Array.from(new Set(
					section.contributions
						.filter(contribution => !contribution.unlist)
						.map(contribution => contribution.author)
				));
				for (const contributor of contributors) {
					$gameMessage.add("- " + contributor);
				}
				$gameMessage.add("");
			}
			interpreter.setWaitMode("message");
		}).finally(() => $gameMessage._isLoadingCredits = false);
	});

})();