//=============================================================================
// RPG Maker MZ - Extra Autosave Conditions
// ----------------------------------------------------------------------------
// (C)2021 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2021/12/16
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Extra Autosave Conditions
 * @author Gaurav Munjal
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help extra_autosave_conditions.js
 *
 * This plugin adds additional auto save conditions and makes the existing ones configurable
 * by the user in the options menu. Also provides some utility commands to trigger autosaves.
 *
 * @param autosavePopupLabel_cte
 * @type text
 * @text Autosave Popup Label
 * @desc The text displayed in the popup when an autosave occurs.
 * @default Autosaved.
 * 
 * @param autosaveCommand_cte
 * @type text
 * @text Autosave Command
 * @desc The text displayed for the Autosave command in options.
 * @default Autosave
 * 
 * @param menuExitLabel_cte
 * @type text
 * @text Menu Exit Label
 * @desc The text used for the menu exit option.
 * @default Menu Exit
 * 
 * @param customAutosaveTriggerLabel_cte
 * @type text
 * @desc The text displayed in the options menu for the custom autosave trigger condition.
 * @default Day Start
 * 
 * @param battleWonLabel_cte
 * @type text
 * @text Battle Won Label
 * @desc The text used for the battle won option.
 * @default Battle Won
 * 
 * @param mapChangeLabel_cte
 * @type text
 * @text Map Change Label
 * @desc The text used for the map change option.
 * @default Map Change
 * 
 * @param offLabel_cte
 * @type text
 * @text Off Label
 * @desc The text used when all options are disabled.
 * @default OFF
 * 
 * @param defaultLabel_cte
 * @type text
 * @text Default Label
 * @desc The text used for the default options.
 * @default Default
 * 
 * @param customLabel_cte
 * @type text
 * @text Custom Label
 * @desc The text used for the custom option.
 * @default Custom
 *
 * @command autosaveNow
 * @text Autosave now
 * @desc Autosave now
 *
 * @command autosave_menu
 * @text Autosave Menu
 * @desc Autosave Menu
 *
 * @command customAutosaveTrigger
 * @text Custom Autosave Trigger
 * @desc Trigger an autosave condition. The player can turn it on/off in the options menu.
 *
 * Dependencies:
 * - Text_Popup.js
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.ExtraAutosave = {};

(($) => {
  const PLUGIN_NAME = "extra_autosave_conditions";
  const PARAMS = PluginManager.parameters(PLUGIN_NAME);
  
  Object.defineProperties($, {
    "autosavePopupLabel": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-autosavePopupLabel_cte`) || PARAMS.autosavePopupLabel_cte },
    "autosaveCommand": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-autosaveCommand_cte`) || PARAMS.autosaveCommand_cte },
    "menuExitLabel": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-menuExitLabel_cte`) || PARAMS.menuExitLabel_cte },
    "customAutosaveTriggerLabel": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-customAutosaveTriggerLabel_cte`) || PARAMS.customAutosaveTriggerLabel_cte },
    "battleWonLabel": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-battleWonLabel_cte`) || PARAMS.battleWonLabel_cte },
    "mapChangeLabel": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-mapChangeLabel_cte`) || PARAMS.mapChangeLabel_cte },
    "offLabel": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-offLabel_cte`) || PARAMS.offLabel_cte },
    "defaultLabel": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-defaultLabel_cte`) || PARAMS.defaultLabel_cte },
    "customLabel": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-customLabel_cte`) || PARAMS.customLabel_cte }
	});

	// Registers a command to Autosave now
	PluginManager.registerCommand(PLUGIN_NAME, "autosaveNow", () => {
		SceneManager._scene.requestAutosave();
	});

    // Registers a command to show the autosave submenu
	PluginManager.registerCommand(PLUGIN_NAME, "autosave_menu", () => {
        SceneManager.push(Scene_Autosave_Menu);
	});

    // Registers a command to indicate the custom trigger condition occurred
	PluginManager.registerCommand(PLUGIN_NAME, "customAutosaveTrigger", () => {
        if (shouldAutosaveOnCustomAutosaveTrigger()) {
            SceneManager._scene.requestAutosave();
        }
	});

    // Show a text popup on autosave success
    const _Scene_Base_onAutoSaveSuccess = Scene_Base.prototype.onAutosaveSuccess;
    Scene_Base.prototype.onAutosaveSuccess = function() {
        _Scene_Base_onAutoSaveSuccess.call(this);
        setTimeout(() => {
            SceneManager.callPopup($.autosavePopupLabel, 'topLeft', 200);
        }, 1000);
    };

    // add autosave option to General Options
	const _Window_Options_addGeneralOptions = Window_Options.prototype.addGeneralOptions;
	Window_Options.prototype.addGeneralOptions = function() {
		_Window_Options_addGeneralOptions.call(this);
		this.addCommand($.autosaveCommand, "autosave_menu");
	};

    Window_Options.prototype.isAutosaveSymbol = function(symbol) {
        return symbol === 'autosave_menu';
    };

    Window_Options.prototype.autosaveSymbol = function() {
        return 'autosave_menu';
    };

	// Override showing value for Autosave
	const _Window_Options_statusText = Window_Options.prototype.statusText;
	Window_Options.prototype.statusText = function(index) {
		const symbol = this.commandSymbol(index);
		if (this.isAutosaveSymbol(symbol)) {
			return this.autosaveStatusText(index);
		} else {
			return _Window_Options_statusText.call(this, index);
		}
	};

	ConfigManager.autosaveMenuExit = false;
	ConfigManager.autosaveCustomTrigger = true;
	ConfigManager.autosaveBattleWon = false;
	ConfigManager.autosaveMapChange = true;

    Window_Options.prototype.autosaveStatusText = function(_) {
        const config = ConfigManager;

        const flags = (
            (config.autosaveMenuExit ? 0x8 : 0) |
            (config.autosaveCustomTrigger ? 0x4 : 0) |
            (config.autosaveBattleWon ? 0x2 : 0) |
            (config.autosaveMapChange ? 0x1 : 0)
        );

        // display as enabled (in white) if not off.
        ConfigManager[this.autosaveSymbol()] = flags !== 0;

        switch (flags) {
            case 0x8: return $.menuExitLabel;
            case 0x4: return $.customAutosaveTriggerLabel;
            case 0x2: return $.battleWonLabel;
            case 0x1: return $.mapChangeLabel;
            case 0: return $.offLabel;
            case 0x5: return $.defaultLabel;
            default: return $.customLabel;
        }
    };

	// Inject logic for handling ok press on autosave menu
	const _Window_Options_processOk = Window_Options.prototype.processOk;
	Window_Options.prototype.processOk = function() {
		const index = this.index();
		const symbol = this.commandSymbol(index);
		if (this.isAutosaveSymbol(symbol)) {
            SceneManager.snapForBackground();
            SceneManager.push(Scene_Autosave_Menu);
		} else {
			_Window_Options_processOk.call(this);
		}
	};

    const ConfigManager_makeData = ConfigManager.makeData;
    ConfigManager.makeData = function() {
        const config = ConfigManager_makeData.call(this);
        config.autosaveCustomTrigger = this.autosaveCustomTrigger;
        config.autosaveMapChange = this.autosaveMapChange;
        config.autosaveBattleWon = this.autosaveBattleWon;
        config.autosaveMenuExit = this.autosaveMenuExit;
        return config;
    };

    const ConfigManager_applyData = ConfigManager.applyData;
    ConfigManager.applyData = function(config) {
        ConfigManager_applyData.call(this, config);
        this.autosaveCustomTrigger = this.readFlag(config, "autosaveCustomTrigger", true);
        this.autosaveMapChange = this.readFlag(config, "autosaveMapChange", true);
        this.autosaveBattleWon = this.readFlag(config, "autosaveBattleWon", false);
        this.autosaveMenuExit = this.readFlag(config, "autosaveMenuExit", false);
    };

    class AutosaveWindow extends Window_Options {
        constructor(rect) {
            super(rect);
        }

        makeCommandList() {
            this.addCommand($.customAutosaveTriggerLabel, 'autosaveCustomTrigger');
            this.addCommand($.mapChangeLabel, 'autosaveMapChange');
            this.addCommand($.battleWonLabel, 'autosaveBattleWon');
            this.addCommand($.menuExitLabel, 'autosaveMenuExit');
        }

        static autosaveWindowRect() {
            const ww = 424;
            const wh = 483;
            const wx = (Graphics.boxWidth - ww) / 2;
            const wy = (Graphics.boxHeight - wh) / 2;
            return new Rectangle(wx, wy, ww, wh);
        }
    }
	
	const _Scene_MenuBase_createBackground = Scene_MenuBase.prototype.createBackground;
    class Scene_Autosave_Menu extends Scene_MenuBase {
        constructor(rect) {
            super(rect);
        }

        create() {
            super.create();
            this.createBackground();
            this.createWindowLayer();
            const rect = AutosaveWindow.autosaveWindowRect();
            this._window = new AutosaveWindow(rect)
            this.addWindow(this._window);
        }

        start() {
            super.start();
            this._window.setHandler('ok', this.processOk.bind(this));
            this._window.setHandler('cancel', this.onCancel.bind(this));
            this._window.activate();
        }

        createBackground() {
            _Scene_MenuBase_createBackground.call(this);
        }

        processOk() {
            const index = this.index();
            const symbol = this.commandSymbol(index);
            this._window.changeValue(symbol, !this.getConfigValue(symbol));
        }

        onCancel() {
            this._window.deactivate();
            this._window.close();
            SceneManager.pop();
        }
    }

    const _Scene_Map_shouldAutosave = Scene_Map.prototype.shouldAutosave;
    Scene_Map.prototype.shouldAutosave = function() {
        return ConfigManager.autosaveMapChange &&
            _Scene_Map_shouldAutosave.call(this);
    };

	Scene_Battle.prototype.shouldAutosave = function() {
		return ConfigManager.autosaveBattleWon &&
			SceneManager.isNextScene(Scene_Map) &&
			BattleManager.result === 0; // win
	}

    const _Scene_Menu_terminate = Scene_Menu.prototype.terminate;
    Scene_Menu.prototype.terminate = function() {
        _Scene_Menu_terminate.call(this);
        if (shouldAutosaveOnMenuExit()) {
            SceneManager._scene.requestAutosave();
        }
    }

    function shouldAutosaveOnCustomAutosaveTrigger() {
        return ConfigManager.autosaveCustomTrigger;
    }

    function shouldAutosaveOnMenuExit() {
        return ConfigManager.autosaveMenuExit;
    }
})(AuraMZ.ExtraAutosave);
