//=============================================================================
// RPG Maker MZ - Show Choices to bypass obstacles with items/skills
// ----------------------------------------------------------------------------
// (C)2021 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2022/01/01
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Show list of choices for items/skills to bypass obstacles
 * @author Gaurav Munjal
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help obstacles_choices.js
 *
 * This plugin allows to show choices for bypassing obstacles with items
 * or skills based on notetags of the form <burn:x> or <explode:x>.
 *
 * Dependencies:
 * - event_utils
 *
 * @command showChociesForObstacle
 * @text Show Choices For Obstacle
 * @desc Show Choices For Obstacle
 *
 * @arg obstacleType
 * @type string
 * @text Obstacle Type
 * @desc Obstacle type, either 'burn' or 'explode'
 *
 * @arg text
 * @type string
 * @text Text to Display
 * @desc Text to Display while showing choices
 *
 * @arg level
 * @type number
 * @text Level of Obstacle
 * @desc Level of Obstacle.
 * @default 1
 *
 * @arg commonEvent
 * @type common_event
 * @text Common Event for animation
 * @desc Common Event for animation
 *
 * @arg scriptOption
 * @type boolean
 * @text Should Show Choice For Script Option
 * @desc Should Show Choice For Script Option
 *
 * @arg variableForReturnValue
 * @type variable
 * @text Variable to set to 0 or 1 for accepted/cancel
 * @desc Variable to set to 0 if user cancelled, 1 if user accepted
 * 
 * @param pluralizeFunc_cte
 * @type text
 * @text Pluralize Function
 * @desc The function to use to pluralise a word.
 * @default (word, num) => word + ((num === 1 || !word.slice(-1).match(/[a-z]/i)) ? "" : "s");
 * 
 * @param capitalizeFunc_cte
 * @type text
 * @text Capitalize Function
 * @desc The function to use to capitalise a word.
 * @default str => str.replace(/\b\w/g, c => c.toUpperCase());
 * 
 * @param scriptOption
 * @type note
 * @text Script Option
 * @desc An option that executes a script if chosen
 * 
 * @param scriptOptionText_cte
 * @type text
 * @text Script Option Text
 * @desc The text to use for the script option
 * @default Remove Collar. (+\V[568] Corruption)
 * @parent scriptOption
 * 
 * @param scriptOptionCondition
 * @type text
 * @text Script Option Condition
 * @desc The condition that needs to be true to show the script condition
 * @default true
 * @parent scriptOption
 * 
 * @param useSkillText_cte
 * @type text
 * @text Use Skill Text
 * @desc The text to use for the Use Skill choice.
 * @default `${actor.name()} Use ${skill.name}. (-${actor.actorSkillCost(skill).value} HP)`
 * 
 * @param castSkillText_cte
 * @type text
 * @text Cast Skill Text
 * @desc The text to use for the Cast Skill choice.
 * @default `${actor.name()} Cast ${skill.name}. (-${actor.skillMpCost(skill)} MP)`
 * 
 * @param useItemText_cte
 * @type text
 * @text Use Item Text
 * @desc The text to use for the Use Item choice.
 * @default `Use ${eval($.capitalize)(item.name)}. (-${num} ${eval($.pluralize)(item.name, num)})`
 * 
 * @param leaveText_cte
 * @type text
 * @text Leave Text
 * @desc The text to use for the Leave choice.
 * @default Leave.
 * 
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.ObstaclesChoices = {};

(($) => {
  const PLUGIN_NAME = "obstacles_choices"
  const PARAMS = PluginManager.parameters(PLUGIN_NAME);
  $.scriptOptionScript = new Function(PARAMS.scriptOption !== "" ? JSON.parse(PARAMS.scriptOption) : "");
  $.scriptOptionCondition = PARAMS.scriptOptionCondition;
  
  Object.defineProperties($, {
    "pluralize": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-pluralizeFunc_cte`) || PARAMS.pluralizeFunc_cte },
    "capitalize": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-capitalizeFunc_cte`) || PARAMS.capitalizeFunc_cte },
    "scriptOption": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-scriptOptionText_cte`) || PARAMS.scriptOptionText_cte },
    "useSkill": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-useSkillText_cte`) || PARAMS.useSkillText_cte },
    "castSkill": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-castSkillText_cte`) || PARAMS.castSkillText_cte },
    "useItem": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-useItemText_cte`) || PARAMS.useItemText_cte },
    "leave": { get: () => TextManager.message(`auramz/${PLUGIN_NAME}-leaveText_cte`) || PARAMS.leaveText_cte }
  });
  
	function getUsablePartySkills(obstacleType, level) {
		const partySkills = [];
		for (const actor of $gameParty.members()) {
			for (const skill of actor.skills()) {
				if (isUseableSkill(actor, skill, obstacleType, level)) {
					partySkills.push({ actor, skill });
				}
			}
		}
		return partySkills;
	}
	
	// Gets all items in possession fo the party that can overcome an obstacle
	function getUsablePartyItems(obstacleType, level) {
		const items = $gameParty.items();
		return items.filter(item => isUseableItem(item, obstacleType, level));
	}

	Game_Message.prototype.useSkill = function() {
		const interpreter = $gameMap.getCurrentInterpreter();
		interpreter.command117([this.commonEvent]);

		const n = $gameMessage.choiceId;
		const actor = $gameActors.actor(this.actorIds[n]);
		actor.paySkillCost($dataSkills[this.skillIds[n]]);
		this.setReturnVariable(1);
	}

	Game_Message.prototype.useItem = function() {
		const interpreter = $gameMap.getCurrentInterpreter();
		interpreter.command117([this.commonEvent]);

		const n = $gameMessage.choiceId;
		$gameParty.loseItem($dataItems[this.itemIds[n]], this.nums[n]);
		this.setReturnVariable(1);
	}

	Game_Message.prototype.scriptOption = function() {
		const interpreter = $gameMap.getCurrentInterpreter();
		$.scriptOptionScript.call(interpreter);
		this.setReturnVariable(1);
	}

	Game_Message.prototype.setReturnVariable = function(value) {
		if (this.returnVariable) {
			$gameVariables.setValue(this.returnVariable, value);
		}
	}

	Game_Message.prototype.createChoiceCallback = function(choices, commonEvent, returnVariable) {
		this.callbacks = [];
		let index = 0;

		// assign callback functions
		for (const choice of choices) {
			if (choice.type === 'scriptOption') {
				this.callbacks.push(() => $gameMessage.scriptOption());
			}
			if (choice.type === 'useItem') {
				this.itemIds = this.itemIds || {};
				this.itemIds[index] = choice.item.id;
				this.nums = this.nums || {};
				this.nums[index] = choice.num;
				this.commonEvent = commonEvent;
				this.returnVariable = returnVariable;
				this.callbacks.push(() => $gameMessage.useItem());
			}
			if (choice.type === 'castSkill') {
				this.actorIds = this.actorIds || {};
				this.actorIds[index] = choice.actor.actorId();
				this.skillIds = this.skillIds || {};
				this.skillIds[index] = choice.skill.id;
				this.commonEvent = commonEvent;
				this.returnVariable = returnVariable;
				this.callbacks.push(() => $gameMessage.useSkill());
			}
			if (choice.type === 'leave') {
				this.callbacks.push(() => $gameMessage.setReturnVariable(0));
			}
			index++;
		}

		// return a function that calls the appropriate callback
		return n => {
			$gameMessage.choiceId = n;
			$gameMessage.callbacks[n]();
			delete $gameMessage.choiceId;
			delete $gameMessage.actorIds;
			delete $gameMessage.skillIds;
			delete $gameMessage.itemIds;
			delete $gameMessage.nums;
			delete $gameMessage.commonEvent;
			delete $gameMessage.returnVariable;
			delete $gameMessage.callbacks;
		}
	}

	function setChoicesForObstacle(obstacleType, level, commonEvent, scriptOption, returnVariable) {
		const usableItems = getUsablePartyItems(obstacleType, level);
		const usablePartySkills = getUsablePartySkills(obstacleType, level);
		const choices = [];

		// If no item or party skill is available
		// search for the lowest level / lowest id item/skill that can be used to bypass the obstacle
		if (usableItems.length == 0 && usablePartySkills.length == 0) {
			getUsableItems($dataItems, obstacleType, level, usableItems);

			if (usableItems.length == 0) {
				getUsableSkills(obstacleType, level, usablePartySkills);
			}
		}

		if (scriptOption) {
			choices.push({
				type: 'scriptOption',
				text: $.scriptOption,
				condition: eval($.scriptOptionCondition)
			});
		}

		for (const partySkill of usablePartySkills) {
			const { actor, skill } = partySkill;
						
			choices.push({
				type: 'castSkill',
				actor,
				skill,
				text: skill.hitType == 1 ? eval($.useSkill) : eval($.castSkill),
				condition:	`const actor = $gameActors.actor(${actor._actorId}); const skill = $dataSkills[${skill.id}]; actor.hasSkill(${skill.id}) ` +
							`&& actor.canPaySkillCost(skill) && !actor.isSkillSealed(${skill.id}) && !actor.isSkillTypeSealed(skill.stypeId)`
			});
		}

		for (const item of usableItems) {
			const num = Math.ceil(level / eval(item.meta[obstacleType]));
			choices.push({
				type: 'useItem',
				item,
				num,
				text: eval($.useItem),
				condition: "$gameParty.numItems($dataItems[" + item.id + "]) >= " + num
			})
		}

		choices.push({
			type: 'leave',
			text: $.leave
		});

		const lastChoice = choices.length - 1;
		$gameMessage.setChoices(choices.map(choice => choice.text), lastChoice, lastChoice);
		for (let i = 0; i < choices.length; ++i) {
			$gameMessage._enableConditions[i] = choices[i].condition;
		}
		$gameMessage.setChoiceCallback($gameMessage.createChoiceCallback(choices, commonEvent, returnVariable));
	}
	
	function isUseableItem(item, obstacleType, level) {
		return item && item.meta[obstacleType] && Math.ceil(level / eval(item.meta[obstacleType])) <= $gameParty.maxItems();
	}
	
	function isUseableSkill(actor, skill, obstacleType, level) {
		return skill && eval(skill.meta[obstacleType]) >= level && actor.skillTypes().includes(skill.stypeId);
	}
	
	// Gets the first item in the database that can overcome an obstacle
	function getUsableItems(itemList, obstacleType, level, currentItems) {
		const item = itemList.find(item => isUseableItem(item, obstacleType, level));
		if (item) {
			currentItems.push(item);
		}
	}
	
	// Gets the first skill in the database the party leader can use to overcome the obstacle
	function getUsableSkills(obstacleType, level, currentSkills) {
		const actor = $gameParty.leader();
		const skill = $dataSkills.find(skill => isUseableSkill(actor, skill, obstacleType, level));
		if (skill) {
			currentSkills.push({ actor, skill });
		}
	}

	PluginManager.registerCommand(PLUGIN_NAME, "showChociesForObstacle", args => {
		const obstacleType = args.obstacleType;
		const text = args.text;
		const level = parseInt(args.level || 1); // do not allow level 0
		const commonEvent = parseInt(args.commonEvent);
		const scriptOption = args.scriptOption === 'true';

		let returnVariable = args.variableForReturnValue;
		returnVariable = returnVariable ? parseInt(returnVariable) : null;

		setChoicesForObstacle(obstacleType, level, commonEvent, scriptOption, returnVariable);
		$gameMessage.setSpeakerName($gameParty.leader().name());
		$gameMessage.add(text.replace(/\\n/g, "\n"));

		const interpreter = $gameMap.getCurrentInterpreter();
		interpreter.setWaitMode('message');
	});

})(AuraMZ.ObstaclesChoices);
