//=============================================================================
// RPG Maker MZ - Skill Rank
// ----------------------------------------------------------------------------
// (C) 2025 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2025/02/12
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Skill Rank Tag
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help skill_rank.js
 *
 * This plugin adds a tag for associating skills with ranks.
 * Ranked skills must be sequentially aligned in the database
 * and have a marker skill at the beginning.
 * 
 * Executes the note in a learning of a class.
 * Use this to remove lower ranks of a skill.
 * 
 * Example:
 * 1 --- MySkill
 * 2 MySkill I <rank:1>
 * 3 MySkill II <rank:2>
 * 4 MySkill III <rank:3>
 * 
 * Added tags:
 * <rank: NUMBER>
 * - Rank of the skill in between 1 - maxSkillRank.
 * 
 * @param maxSkillRank
 * @text Max Skill Rank
 * @type number
 * @desc Maximum rank of a skill. The number of slots that must be kept free after the base skill in the database.
 * @default 3
 */

window.AuraMZ = window.AuraMZ || {};
AuraMZ.SkillRank = {};

(($) => {
	const PLUGIN_ID = "skill_rank";
	$.Params = PluginManager.parameters(PLUGIN_ID);
	$.maxSkillRank = parseInt($.Params["maxSkillRank"]);

	// Gets the rank of a skill. Input ID must be of the skill before rank I.
	Game_BattlerBase.prototype.getSkillRank = function(index) {
		const skill = this.skills().find(skill => skill.id > index && skill.id <= index + $.maxSkillRank);
		if (skill) {
			return skill.id - index;
		} else {
			return 0;
		}
	}

	// Learns the skill or increases its rank by 1. Input ID must be of the skill before rank I.
	Game_BattlerBase.prototype.increaseSkillRank = function(index) {
		const skillRank = this.getSkillRank(index);
		if (skillRank < $.maxSkillRank) {
			this.learnSkill(index + skillRank + 1);
			if (skillRank > 0) {
				this.forgetSkill(index + skillRank);
			}
			return true;
		}
		
		return false;
	};
	
	// Forgets the skill or decreases its rank by 1. Input ID must be of the skill before rank I.
	Game_BattlerBase.prototype.decreaseSkillRank = function(index) {
		const skillRank = this.getSkillRank(index);
		if (skillRank > 0) {
			if (skillRank > 1) {
				this.learnSkill(index + skillRank - 1);
			}
			this.forgetSkill(index + skillRank);
			return true;
		}
		
		return false;
	};

	// Gets the base skill for a ranked skill
	Game_BattlerBase.prototype.getBaseSkillId = function(rankedSkillId) {
		return rankedSkillId - parseInt($dataSkills[rankedSkillId].meta.rank);
	}

	// Inject custom level up logic to execute JS from the note of a learned skill
	// and make sure to also learn lower level spells for easier migration
	Game_Actor.prototype.levelUp = function() {
		this._level++;
		for (const learning of this.currentClass().learnings) {
			if (learning.level <= this._level) {
				if ($dataSkills[learning.skillId].meta.rank) {
					const baseSkill = this.getBaseSkillId(learning.skillId);
					const currentRank = this.getSkillRank(baseSkill);
					if (currentRank > $dataSkills[learning.skillId].meta.rank) continue;
				}
				this.applyLearning(learning);
			}
		}
	};

	// Inject custom logic to evaluate the learning note when applying a learning to an actor
	Game_Actor.prototype.applyLearning = function(learning) {
		this.learnSkill(learning.skillId);
		if (learning.note) {
			eval(learning.note);
		}
	}

	// Inject custom logic that if skillUses is being tracked, higher ranked skills
	// carry over the skill uses. Requires TLB_SKABattleHUD.js
	const Game_Actor_learnSkill = Game_Actor.prototype.learnSkill;
	Game_Actor.prototype.learnSkill = function(skillId) {
		Game_Actor_learnSkill.call(this, skillId);
		const skill = $dataSkills[skillId];
		if (skill.meta.rank
			&& parseInt(skill.meta.rank) >= 2
			&& !this._skillUses?.[skillId]
			&& this._skillUses?.[skillId - 1]) {
			this._skillUses[skillId] = this._skillUses[skillId - 1];
		}
	}
})(AuraMZ.SkillRank);

