//=============================================================================
// RPG Maker MZ - Standing Images
// ----------------------------------------------------------------------------
// (C)2021 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 2021/08/02
// 1.1.0 2022/02/19 - Added offset parameter
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Standing Images
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help standing_images.js
 *
 * This plugin provides services for displaying standing images of characters.
 *
 * @command show
 * @text Show Standing Image
 * @desc Shows a standing image
 *
 * @arg parts
 * @type struct<Part>[]
 * @text body
 * @desc The name of the body / pose image
 *
 * @arg side
 * @type text
 * @text side
 * @desc The side (0 = left, 1 = right, 2 = middle) where the standing image will appear.
 * @default 0
 *
 * @arg fadeIn
 * @type text
 * @text fadeIn
 * @desc The number of fade in frames
 * @default 10
 *
 * @arg scale
 * @type text
 * @text scale
 * @desc How much the standing image will be scaled
 * @default 50
 *
 * @arg xOffset
 * @type text
 * @text xOffset
 * @desc How much the standimg image will be offset in the x axis
 * @default 0
 *
 * @arg yOffset
 * @type text
 * @text yOffset
 * @desc How much the standimg image will be offset in the y axis
 * @default 0
 *
 * @command hide
 * @text Hide Standing Image
 * @desc Hide a standing image
 *
 * @arg side
 * @type text
 * @text side
 * @desc The side (0 = left, 1 = right) which standing image will disappear.
 * @default 0
 *
 * @arg fadeOut
 * @type text
 * @text fadeOut
 * @desc The number of fade out frames
 * @default 5
 *
 * @command setNameMap
 * @text Set Name Map
 * @desc Sets a name map
 *
 * @arg nameMap
 * @type struct<NameMap>[]
 * @text Name Map
 * @desc Re-Maps the names for identifying assets
 *
 * @param Name Mappings
 * @type struct<NameMap>[]
 * @desc Maps a name to another name for associating a speaker with an asset
 *
 * @command lockDim
 * @text Lock Dim
 * @desc Locks the dimming on the specified side
 *
 * @arg side
 * @type text
 * @text side
 * @desc The side (0 = left, 1 = right) which standing image will disappear.
 * @default 0
 *
 */

/*~struct~Part:
 *
 * @param imageName
 * @type string
 * @text imageName
 * @desc The name of the image of this part
 *
 * @param condition
 * @type string
 * @text condition
 * @desc The condition for the part to be shown
 *
 */

/*~struct~NameMap:
 *
 * @param name
 * @type string
 * @text Name
 * @desc A message box name
 *
 * @param mappedName
 * @type string
 * @text Mapped Name
 * @desc The name the message box name will be mapped to
 */


class StandingImageUtil {
	static getX(side) {
		if ((side / 10) % 2 == 1) {
			return 0;
		} else {
			return Graphics.width;
		}
	}
}

window.AuraMZ = window.AuraMZ || {};
AuraMZ.StandingImages = {};

(($) => {
	$.PLUGIN_NAME = "standing_images";
	$.PARAMS = PluginManager.parameters($.PLUGIN_NAME);
	$.PARTS_PER_SIDE = 10;
	$.STANDING_IMAGE_PREFIX = "si/";
	$.EASING_MODE = 0;
	$.TINT_DURATION = 30;
	$.NON_SPEAKER_TINT = [-30, -30, -30, 120];
	$.SPEAKER_TINT = [0, 0, 0, 0];

	let eraseSide = 0;
	let finishedShowCommand = false;
	let fadeIn = 0;
	let bitmaps = [];
	$.nameMappings = [];

	function show(args) {
		const parts = JSON.parse(args.parts);
		fadeIn = eval(args.fadeIn);
		const side = (eval(args.side) + 1) * $.PARTS_PER_SIDE;

		finishedShowCommand = false;
		bitmaps = [];

		const processedParts = parts.map(part => {
			let { imageName, condition } = JSON.parse(part);
			imageName = eval(imageName);
			condition = eval(condition);
			return { imageName, condition };
		});

		Promise.all(processedParts.map(part => createBmpPromise(part))).then(() => {
			showProcessedParts(processedParts, args.xOffset, args.yOffset, args.scale, side);

			// Erase old parts
			for (let i = parts.length; i < $.PARTS_PER_SIDE; ++i) {
				$gameScreen.erasePicture(side + i);
			}

			finishedShowCommand = true;
		});
		
		this.setWaitMode("standing-image");
	}

	// Command to show a standing image
	PluginManager.registerCommand($.PLUGIN_NAME, "show", show);

	function createBmpPromise(part) {
		const { imageName, condition } = part;
		if (condition) {
			return new Promise(resolve => {
				const bmap = ImageManager.loadPicture(imageName);
				bitmaps.push(bmap);
				bmap.addLoadListener(resolve);
			});
		} else {
			return Promise.resolve();
		}
	}

	function showProcessedParts(processedParts, xOffsetArg, yOffsetArg, scaleArg, side) {
		const scale = eval(scaleArg);
		const xOffset = xOffsetArg ? eval(xOffsetArg) * scale / 100 : 0;
		const yOffset = yOffsetArg ? eval(yOffsetArg) * scale / 100 : 0;

		const blendMode = 0;

		const maxOpacity = 255;
		const initOpacity = fadeIn == 0 ? maxOpacity : 0;
		const directionScale = ((side / $.PARTS_PER_SIDE) % 2 == 1) ? 1 : -1;
		const origin = 0;

		const padding = 50 - scale;
		const x = directionScale * (padding + xOffset) + StandingImageUtil.getX(side);
		const y = padding + yOffset;

		for (let i = 0; i < processedParts.length; ++i) {
			const part = processedParts[i];
			const pictureID = side + i;
			if (part.condition) {
				const picture = $gameScreen.picture(pictureID);
				const opacity = picture ? maxOpacity : initOpacity;
				$gameScreen.showPicture(
					pictureID, part.imageName, origin, x, y, directionScale * scale, scale, opacity, blendMode
				);

				if (!picture) {
					$.markSpeakerPart(pictureID);
					if (fadeIn != 0) {
						$gameScreen.movePicture(
							pictureID, origin, x, y, directionScale * scale, scale, maxOpacity, blendMode, fadeIn, $.EASING_MODE
						);
					}
				}
			} else {
				$gameScreen.erasePicture(pictureID);
			}
		}
	}

	const _Game_Interpreter_updateWaitMode = Game_Interpreter.prototype.updateWaitMode;
	Game_Interpreter.prototype.updateWaitMode = function() {
		if (this._waitMode == "standing-image") {
			if (finishedShowCommand) {
				this.wait(fadeIn);
				this._waitMode = "";
			} else {
				// If we are still waiting for the bitmaps to finish loading
				// check if there were any loading errors and throw an exception
				// if that is the case. The isReady method of ImageManager does both
				for (const bitmap of bitmaps) {
					ImageManager.isReady(bitmap);
				}
			}

			return !finishedShowCommand;
		} else {
			return _Game_Interpreter_updateWaitMode.call(this);
		}
	};

	function hide(args) {
		const fadeOut = args.fadeOut;
		eraseSide = (eval(args.side) + 1) * $.PARTS_PER_SIDE;

		for (let i = 0; i < $.PARTS_PER_SIDE; ++i) {
			const pictureID = eraseSide + i;
			const picture = $gameScreen.picture(pictureID);

			if (picture) {
				$gameScreen.movePicture(
					pictureID, picture.origin(),
					picture.x(), picture.y(),
					picture.scaleX(), picture.scaleY(),
					0, picture.blendMode(),
					fadeOut, $.EASING_MODE
				);
			}
		}

		// Wait until fadeout complete
		if ($gameMap.event(this._eventId)?._trigger != 4) {
			// Don't wait in parallel events
			this.wait(fadeOut);
			this.setWaitMode("fadeOut");
		}
	}

	// Command to hide a standing image
	PluginManager.registerCommand($.PLUGIN_NAME, "hide", hide);

	// Inject logic to erase standing image after fadeout
	const _Game_Interpreter_updateWaitCount = Game_Interpreter.prototype.updateWaitCount;
	Game_Interpreter.prototype.updateWaitCount = function() {
		const res = _Game_Interpreter_updateWaitCount.call(this);
		if (!res && eraseSide != 0 && this._waitMode == "fadeOut") {
			for (let i = 0; i < $.PARTS_PER_SIDE; ++i) {
				$gameScreen.erasePicture(eraseSide + i);
			}

			for (let i = $.PARTS_PER_SIDE; i < $.PARTS_PER_SIDE * 3; ++i) {
				const picture = $gameScreen.picture(i);
				const pictureName = picture?.name();
				if (pictureName?.startsWith($.STANDING_IMAGE_PREFIX)) {
					$gameScreen.tintPicture(
						i, $.SPEAKER_TINT, $.TINT_DURATION
					);
				}
			}

			this._waitMode = "";
			eraseSide = 0;
		}
		return res;
	}

	const _Game_Message_setSpeakerName = Game_Message.prototype.setSpeakerName;
	Game_Message.prototype.setSpeakerName = function(speakerName) {
		const speakerNameList = $.getSpeakerNameList(speakerName);

		for (let i = $.PARTS_PER_SIDE; i < $.PARTS_PER_SIDE * 3; ++i) {
			if ($.lockDim != undefined) {
				const side = Math.floor((i - $.PARTS_PER_SIDE) / $.PARTS_PER_SIDE);
				if (side == $.lockDim) {
					continue;
				}
			}

			const picture = $gameScreen.picture(i);
			const pictureName = picture?.name();
			if (pictureName?.startsWith($.STANDING_IMAGE_PREFIX)) {
				const isSpeaker = $.isSpeaker(pictureName, speakerNameList);
				const onScreenSpeakers = $.getOnScreenSpeakers();
				const isOffscreenSpeaker = onScreenSpeakers.filter(speaker => speakerNameList.contains(speaker)).length == 0;
				if (onScreenSpeakers.length >= 2 && !isSpeaker && !isOffscreenSpeaker) {
					$.markNonSpeakerPart(i);
				} else {
					$.markSpeakerPart(i);
				}
			}
		}

		$.lockDim = undefined;
		_Game_Message_setSpeakerName.call(this, speakerName);
	};

	// Gets the collection of speaker names, applying all substitutions and
	// considering multi speakers separated by ","
	const andRegex = /&/g;
	$.getSpeakerNameList = function(speakerName) {
		return speakerName.toLowerCase().replace(andRegex, ",").split(",")
			.map(name => name.trim())
			.map(name => $.nameMappings[name] ? $.nameMappings[name] : name);
	}

	// Gets the speakers currently displayed on the screen through standing images
	$.getOnScreenSpeakers = function() {
		const onScreenSpeakers = [];

		for (let i = 0; i < 2; ++i) {
			const start = $.PARTS_PER_SIDE * (i + 1);
			const end = start + $.PARTS_PER_SIDE;
			for (let j = start; j < end; ++j) {
				const picture = $gameScreen.picture(j);
				if (picture) {
					const pictureName = picture.name();
					if (pictureName.startsWith($.STANDING_IMAGE_PREFIX)) {
						onScreenSpeakers.push(pictureName.split("/")[1]);
						break;
					}
				}
			}
		}

		return onScreenSpeakers;
	}

	// Checks if the given picture belongs to the speaker
	// It belongs to the speaker if any individual speaker 
	// is contained in the asset name of the picture
	$.isSpeaker = function(pictureName, speakerNameList) {
		return speakerNameList.some(speakerName => pictureName.contains(speakerName));
	}

	$.markSpeakerPart = function(pictureID) {
		const picture = $gameScreen.picture(pictureID);
		if (picture) {
			$gameScreen.tintPicture(
				pictureID, $.SPEAKER_TINT, 0
			);
		}
	}

	$.markNonSpeakerPart = function(pictureID) {
		const picture = $gameScreen.picture(pictureID);
		if (picture) {
			$gameScreen.tintPicture(
				pictureID, $.NON_SPEAKER_TINT, $.TINT_DURATION
			);
		}
	}

	$.setNameMap = function(nameMapString) {
		const NAME_MAP = JSON.parse(nameMapString);

		for (const nameMappingString of NAME_MAP) {
			const nameMapping = JSON.parse(nameMappingString);
			const name = nameMapping["name"];
			const mappedName = nameMapping["mappedName"];
			$.nameMappings[name] = mappedName;
		}
	}

	$.setNameMap($.PARAMS["Name Mappings"]);

	// Command to dim a standing image
	PluginManager.registerCommand($.PLUGIN_NAME, "setNameMap", args => {
		$.setNameMap(args.nameMap);
	});

	// Command to dim a standing image
	PluginManager.registerCommand($.PLUGIN_NAME, "lockDim", args => {
		$.lockDim = eval(args.side);
		const side = (eval($.lockDim) + 1) * $.PARTS_PER_SIDE;

		for (let i = 0; i < $.PARTS_PER_SIDE; ++i) {
			const pictureID = side + i;
			const picture = $gameScreen.picture(pictureID);
			if (picture) {
				$.markNonSpeakerPart(pictureID);
			}
		}
	});
})(AuraMZ.StandingImages);
