//=============================================================================
// RPG Maker MZ - Liora Contract
// ----------------------------------------------------------------------------
// (C)2025 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 03/04/2025
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Liora Contract System
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help liora_contract.js
 *
 * Implements the contract system
 * 
 * @param contractsCommand_cte
 * @text Contracts Command
 * @desc Compendium Command Label for opening the Contracts Category
 * @default Contracts
 * 
 * @param clauseCategories_cte
 * @text Clause Categories
 * @type text[]
 * @desc Categories for contracts
 * @default []
 * 
 * @param triggerClauseText_cte
 * @text Trigger Clause Text
 * @desc Base text for trigger part of a clause
 * @default If \c[10]%1\c[0] %2
 * 
 * @param triggerKillEnemyText_cte
 * @text Trigger Kill Enemy Text
 * @desc Clause text for the Kill Enemy Trigger
 * @default kills an ENEMY
 * @parent triggerClauseText_cte
 * 
 * @param triggerKillBossText_cte
 * @text Trigger Kill Boss Text
 * @desc Clause text for the Kill Boss Trigger
 * @default kills a BOSS
 * @parent triggerClauseText_cte
 * 
 * @param triggerReceivedText_cte
 * @text Trigger Received Text
 * @desc Clause text for receiving a state
 * @default received \I[%2]%1
 * @parent triggerClauseText_cte
 * 
 * @param triggerTurnText_cte
 * @text Trigger Turn Text
 * @desc Clause text for Turn Triggers
 * @default is on TURN %1
 * @parent triggerClauseText_cte
 * 
 * @param triggerDefeatedText_cte
 * @text Trigger Defeated Text
 * @desc Clause text for Defeated Trigger
 * @default is DEFEATED
 * @parent triggerClauseText_cte
 * 
 * @param triggerDialogueTagTexts
 * @type struct<map>[]
 * @text Trigger Dialogue Tags Texts
 * @desc Clause text for dialogue triggers
 * @parent triggerClauseText_cte
 * 
 * @param restrictionText_cte
 * @text Restriction Text
 * @desc Clause text for restrictions
 * @default RESTRICTION:
 * 
 * @param effectClauseText_cte
 * @text Effect Clause Text
 * @desc Base text for effect part of a clause
 * @default then \c[10]%1\c[0] shall:
 * 
 * @param effectClaimLootText_cte
 * @text Effect Claim Loot
 * @desc Clause text for the Claim Loot Effect
 * @default Obtain the right to claim its loot.
 * @parent effectClauseText_cte
 * 
 * @param effectUseSkillText_cte
 * @text Effect Use Skill
 * @desc Clause text for the Claim Loot Effect
 * @default Use %1%2.
 * @parent effectClauseText_cte
 * 
 * @param effectUseSkillRandomText_cte
 * @text Target Random Enemy
 * @desc Target a random enemy for the Use Skill Effect
 * @default on RANDOM enemy
 * @parent effectUseSkillText_cte
 *
 * @param effectUseSkillBossText_cte
 * @text Target Boss Enemy
 * @desc Target a random boss enemy for the Use Skill Effect
 * @default on BOSS enemy
 * @parent effectUseSkillText_cte
 * 
 * @param effectUseSkillActorText_cte
 * @text Target Actor Last Target
 * @desc Target actor's last target for the Use Skill Effect
 * @default on \c[10]%1's\c[0] last target
 * @parent effectUseSkillText_cte
 * 
 * @param effectFlagTexts
 * @type struct<map>[]
 * @text Effect Set Flag Texts
 * @desc Text displayed for setting a flag
 * @parent effectClauseText_cte
 * 
 * @param effectChangeVariableTexts
 * @type struct<map>[]
 * @text Effect Change Variable Texts
 * @desc Text displayed for changing a variable
 * @parent effectClauseText_cte
 * 
 * @param effectWearCostumeText_cte
 * @text Effect Wear Costume Text
 * @desc Clause text for the Wear Costume Effect
 * @default Wear %1 on the following day.
 * @parent effectClauseText_cte
 *
 * @param costumeNames_cte
 * @type text[]
 * @text Costume Names
 * @desc Name of the costumes
 * @default []
 * @parent effectWearCostumeText_cte
 * 
 * @param effectTradeGoldText_cte
 * @text Effect Trade Gold Text
 * @desc Clause text for the Trade Gold Effect
 * @default \c[10]%1\c[0] shall gain %2 Gold.
 * @parent effectClauseText_cte
 * 
 * @command checkDialogueTriggers
 * @text Check Dialogue Triggers
 * @desc Executes a contract check on the dialogue trigger condition
 *
 * @arg subjectId
 * @type number
 * @text Subject Id
 * @desc The subject id of the speaker 
 * @default 1
 * 
 * @arg tags
 * @type text[]
 * @text Tags
 * @desc The list of dialogue tags
 * 
 */
/*~struct~map:
*
* @param key
* @text Key
* @type text
* @desc Key of the map
*
* @param value_cte
* @text Value
* @type text
* @desc The associated text value
*
*/

window.Liora = window.Liora || {};
Liora.Contract = {};

(($) => {
	// ---------------------------------------------------------------------------------
	// DATA
	// ---------------------------------------------------------------------------------
	const PLUGIN_ID = "liora_contract";
	$.Params = PluginManager.parameters(PLUGIN_ID);
	$.compendiumCategory = "contracts";
	$.clauseCategories = JSON.parse($.Params.clauseCategories_cte);
	$.costumeNames = JSON.parse($.Params.costumeNames_cte);
	$.triggerDialogueTagTexts = {};
	$.effectFlagTexts = {};
	$.effectChangeVariableTexts = {};

	Object.defineProperties($, {
		"contractsCommand": { get: () => TextManager.message(`${PLUGIN_ID}-contractsCommand_cte`) || $.Params.contractsCommand_cte },
		"triggerClauseText": { get: () => TextManager.message(`${PLUGIN_ID}-triggerClauseText_cte`) || $.Params.triggerClauseText_cte },
		"triggerKillEnemyText": { get: () => TextManager.message(`${PLUGIN_ID}-triggerKillEnemyText_cte`) || $.Params.triggerKillEnemyText_cte },
		"triggerKillBossText": { get: () => TextManager.message(`${PLUGIN_ID}-triggerKillBossText_cte`) || $.Params.triggerKillBossText_cte },
		"triggerReceivedText": { get: () => TextManager.message(`${PLUGIN_ID}-triggerReceivedText_cte`) || $.Params.triggerReceivedText_cte },
		"triggerTurnText": { get: () => TextManager.message(`${PLUGIN_ID}-triggerTurnText_cte`) || $.Params.triggerTurnText_cte },
		"triggerDefeatedText": { get: () => TextManager.message(`${PLUGIN_ID}-triggerDefeatedText_cte`) || $.Params.triggerDefeatedText_cte },
		"restrictionText": { get: () => TextManager.message(`${PLUGIN_ID}-restrictionText_cte`) || $.Params.restrictionText_cte },
		"effectClauseText": { get: () => TextManager.message(`${PLUGIN_ID}-effectClauseText_cte`) || $.Params.effectClauseText_cte },
		"effectClaimLootText": { get: () => TextManager.message(`${PLUGIN_ID}-effectClaimLootText_cte`) || $.Params.effectClaimLootText_cte },
		"effectUseSkillText": { get: () => TextManager.message(`${PLUGIN_ID}-effectUseSkillText_cte`) || $.Params.effectUseSkillText_cte },
		"effectUseSkillText": { get: () => TextManager.message(`${PLUGIN_ID}-effectUseSkillText_cte`) || $.Params.effectUseSkillText_cte },
		"effectUseSkillRandomText": { get: () => TextManager.message(`${PLUGIN_ID}-effectUseSkillRandomText_cte`) || $.Params.effectUseSkillRandomText_cte },
		"effectUseSkillBossText": { get: () => TextManager.message(`${PLUGIN_ID}-effectUseSkillBossText_cte`) || $.Params.effectUseSkillBossText_cte },
		"effectUseSkillActorText": { get: () => TextManager.message(`${PLUGIN_ID}-effectUseSkillActorText_cte`) || $.Params.effectUseSkillActorText_cte },
		"effectWearCostumeText": { get: () => TextManager.message(`${PLUGIN_ID}-effectWearCostumeText_cte`) || $.Params.effectWearCostumeText_cte },
		"effectTradeGoldText": { get: () => TextManager.message(`${PLUGIN_ID}-effectTradeGoldText_cte`) || $.Params.effectTradeGoldText_cte }
	});

	// Data initialization
	$._Scene_Boot_start = Scene_Boot.prototype.start;
	Scene_Boot.prototype.start = function() {
		$.defaultClauses = [];
		for (const baseClause of $.baseClauses) {
			for (const subjectId of baseClause.subjectIds) {
				for (const condition of baseClause.conditions) {
					for (const effect of baseClause.effects) {
						const defaultClause = {};
						Object.assign(defaultClause, baseClause);
						defaultClause.subjectId = subjectId;
						defaultClause.conditions = [condition];
						defaultClause.effects = [effect];
						defaultClause.minLewdness = defaultClause.minLewdness || (() => -1);
						delete defaultClause.subjectIds;
						$.defaultClauses.push(defaultClause);
					}
				}
			}
		}

		$._Scene_Boot_start.call(this);

		const paramFlagTexts = JSON.parse($.Params["effectFlagTexts"]);
		for (let i = 0; i < paramFlagTexts.length; i++) {
			const flagEntry = JSON.parse(paramFlagTexts[i]);
			$.effectFlagTexts[flagEntry.key] = flagEntry.value_cte;
		}

		const paramChangeVariableTexts = JSON.parse($.Params["effectChangeVariableTexts"]);
		for (let i = 0; i < paramChangeVariableTexts.length; i++) {
			const variableEntry = JSON.parse(paramChangeVariableTexts[i]);
			$.effectChangeVariableTexts[variableEntry.key] = variableEntry.value_cte;
		}

		const paramTagTexts = JSON.parse($.Params["triggerDialogueTagTexts"]);
		for (let i = 0; i < paramTagTexts.length; i++) {
			const tagEntry = JSON.parse(paramTagTexts[i]);
			$.triggerDialogueTagTexts[tagEntry.key] = tagEntry.value_cte;
		}
	};

	// ---------------------------------------------------------------------------------
	// TRIGGERS
	// ---------------------------------------------------------------------------------

	$.PLACEHOLDER = { checkCondition: false, toString: () => "____", execute: () => { }, value: () => 0 };

	$.dataTriggers = [];
	$.dataTriggers[0] = $.PLACEHOLDER;

	$.dataTriggers[1] = {
		name: "KILL_ENEMY",
		checkCondition: (context, clause, condition) => {
			return context.subjectId == clause.clause().subjectId
				&& (condition.params[0] == 0 || context.params[1].enemy().meta.boss);
		},
		toString: (condition) => [$.triggerKillEnemyText, $.triggerKillBossText][condition.params[0]],
		value(clause, condition) {
			const actorValue = this.actorValue(clause.clause().subjectId);
			const enemyValue = this.enemyValue(condition.params[0]);
			return actorValue * enemyValue;
		},
		actorValue(actorId) { return [15, 20, 10][actorId - 1] },
		enemyValue(enemyType) { return [1, 0.5][enemyType] }
	};

	$.dataTriggers[2] = {
		name: "DEFEATED",
		checkCondition(context, clause, _) { return context.subjectId == clause.clause().subjectId },
		toString(_) { return $.triggerDefeatedText },
		value(_, __) { return 20 }
	};

	$.dataTriggers[3] = {
		name: "RECEIVED",
		checkCondition(context, clause, condition) {
			return context.subjectId == clause.clause().subjectId
				&& context.params[0] == condition.params[0]
		},
		toString(condition) {
			const state = $dataStates[condition.params[0]];
			return $.triggerReceivedText.format(state.name.toUpperCase(), state.iconIndex)
		},
		value(clause, condition) {
			const stateValue = this.stateValue(condition.params[0]);
			const actorValue = this.actorValue(clause.clause().subjectId);
			return stateValue * actorValue;
		},
		actorValue(actorId) { return [20, 20, 10][actorId - 1] },
		stateValue(stateId) {
			const state = $dataStates[stateId];
			return parseInt(state.meta.negotiate_value);
		}
	};

	$.dataTriggers[4] = {
		name: "RESTRICT_DAY",
		checkCondition: (_, __, condition) => {
			return Liora.Calendar.currentWeekdayId() == condition.params[0];
		},
		toString: (condition) => {
			const weekday = Liora.Calendar.weekday(condition.params[0]);
			return `${$.restrictionText} \\I[%1]\\c[10]%2\\c[0]`.format(Liora.Calendar.weekdayIconIndex(weekday), weekday.name_cte.toUpperCase());
		},
		value: (_, __) => {
			return 0.5;
		}
	};

	$.dataTriggers[5] = { name: "INSTANT", checkCondition: true, toString: () => "", execute: () => { }, value: () => 1 };

	$.dataTriggers[7] = {
		name: "TURN",
		checkCondition: (_, __, condition) => $gameTroop.turnCount() + 1 == condition.params[0],
		toString: (condition) => $.triggerTurnText.format(condition.params[0]),
		value: (_, condition) => 20 - 2 * condition.params[0]
	};

	$.dataTriggers[8] = {
		name: "DIALOGUE",
		checkCondition: (context, _, condition) => context.tags.includes(condition.params[0]),
		toString: (condition) => $.triggerDialogueTagTexts[condition.params[0]].format(condition.params[1]),
		value: (_, condition) => ({ "joke": 5, "minePyromantium": 40, "gropeDate": 30 }[condition.params[0]]) || 10
	};

	$.TRIGGERS = {};
	$.dataTriggers.forEach((trigger, i) => {
		trigger.id = i;
		$.TRIGGERS[trigger.name] = trigger;
	});

	// ---------------------------------------------------------------------------------
	// EFFECTS
	// ---------------------------------------------------------------------------------

	$.dataEffects = [];
	$.dataEffects[0] = $.PLACEHOLDER;

	$.dataEffects[1] = {
		name: "CLAIM_DROP_ITEMS",
		execute: (context, clause, _) => {
			const enemy = context.params[1];
			if (enemy.makeDropItems().length > 0) {
				enemy._actorClaimDropItems = clause.target().actorId();
			} else {
				context.cannotExecute = true;
			}
		},
		toString: (_) => $.effectClaimLootText,
		value: (_, __) => 10,
		conflict: (_, effect1, effect2) => effect1.id == effect2.id
	};

	$.dataEffects[2] = {
		name: "SET_FLAG",
		execute: (_, __, effect) => $gameSwitches.setValue(effect.params[0], true),
		toString: (effect) => {
			const switchId = effect.params[0];
			const flagText = TextManager.message(`${PLUGIN_ID}-effectFlagTexts-${switchId}-value_cte`) || $.effectFlagTexts[switchId];
			return flagText;
		},
		value: (_, effect) => ({ 202: 10, 203: 15, 204: 40, 205: 10, 206: 40, 207: 40 }[effect.params[0]]),
		conflict: (_, effect1, effect2) => effect1.params[0] == effect2.params[0]
	};

	$.dataEffects[3] = {
		name: "COSTUME",
		execute: (context, _, effect) => {
			if (($gameVariables.value(302) != effect.params[0] && $gameVariables.value(303) == effect.params[0]) || $gameVariables.value(304) > 1) {
				// Already procced, cannot execute
				context.cannotExecute = true;
			} else {
				$gameVariables.setValue(303, effect.params[0]);
				$gameVariables.setValue(304, $gameVariables.value(302) == effect.params[0] ? 2 : 1);
			}
		},
		toString: (effect) => {
			const costumeId = effect.params[0] - 1;
			const costumeName = TextManager.message(`${PLUGIN_ID}-costumeNames_cte-${costumeId}`) || $.costumeNames[costumeId];
			return $.effectWearCostumeText.format(costumeName);
		},
		value: (_, effect) => effect.params[0] * 10,
		conflict: (_, effect1, effect2) => effect1.id == effect2.id
	};

	$.dataEffects[4] = {
		name: "USE_SKILL",
		execute: (_, __, effect) => Liora.Edgar._forceSkill = effect.params,
		toString: (effect) => {
			const skillName = $dataSkills[effect.params[0]].name;
			const target = [
				"",
				` ${$.effectUseSkillRandomText}`,
				` ${$.effectUseSkillBossText}`,
				` ${$.effectUseSkillActorText}`.format($gameActors.actor(1).name().toUpperCase()),
				` ${$.effectUseSkillActorText}`.format($gameActors.actor(2).name().toUpperCase())
			][effect.params[1]];
			return $.effectUseSkillText.format(skillName, target)
		},
		value(_, effect) {
			const skillValue = this.skillValue(effect.params[0]);
			const targetValue = this.targetValue(effect.params[1]);
			return skillValue * targetValue;
		},
		skillValue(skilldId) { return parseInt($dataSkills[skilldId].meta.negotiate_value) },
		targetValue(targetId) { return [1, 2, 0.75, 3, 4][targetId] },
		conflict: (_, effect1, effect2) => effect1.id == effect2.id
	};

	$.dataEffects[5] = {
		name: "TRADE_GOLD",
		execute: (_, __, effect) => {
			const gold = effect.params[1];
			if (effect.params[0] == 1 || effect.params[0] == 2) {
				Liora.Edgar.getParty().loseGold(gold);
				$gameParty.gainGold(gold);
			} else {
				Liora.Edgar.getParty().gainGold(gold);
				$gameParty.loseGold(gold);
			}
		},
		toString: (effect) => {
			return $.effectTradeGoldText.format($gameActors.actor(effect.params[0]).name().toUpperCase(), effect.params[1]);
		},
		value: (_, effect) => {
			return Math.ceil(effect.params[1] / 5);
		},
		conflict: (_, __, ___) => false
	};

	$.dataEffects[6] = {
		name: "CHANGE_VARIABLE",
		execute: (_, __, effect) => $gameVariables.setValue(effect.params[0], $gameVariables.value(effect.params[0]) + effect.params[1]),
		toString: (effect) => {
			const variableId = effect.params[0];
			const change = effect.params[1];
			const changeText = TextManager.message(`${PLUGIN_ID}-effectChangeVariableTexts-${variableId}-value_cte`) || $.effectChangeVariableTexts[variableId];
			return changeText.format(change);
		},
		value: (_, effect) => ({ 602: 600 }[effect.params[0]]),
		conflict: (_, effect1, effect2) => effect1.params[0] == effect2.params[0]
	};

	$.EFFECTS = {};
	$.dataEffects.forEach((effect, i) => {
		effect.id = i;
		$.EFFECTS[effect.name] = effect;
	});

	// ---------------------------------------------------------------------------------
	// BASE CLAUSES FOR EDGAR OFFERS
	// ---------------------------------------------------------------------------------

	$.baseClauses = [
		{
			category: "Combat", subjectIds: [3], targetId: 3,
			conditions: [{ id: $.TRIGGERS.KILL_ENEMY.id, params: [0] }],
			effects: [{ id: $.EFFECTS.CLAIM_DROP_ITEMS.id, params: [] }]
		},
		{
			category: "Social", subjectIds: [3], targetId: 1,
			conditions: [{ id: $.TRIGGERS.DIALOGUE.id, params: ["joke", 1] }],
			effects: [{ id: $.EFFECTS.SET_FLAG.id, params: [202] }],
		},
		{
			category: "Social", subjectIds: [1, 2], targetId: 1,
			conditions: [{ id: $.TRIGGERS.DEFEATED.id, params: [] }],
			effects: [{ id: $.EFFECTS.SET_FLAG.id, params: [203] }, { id: $.EFFECTS.COSTUME.id, params: [1] }],
		},
		{
			category: "Social", subjectIds: [12], targetId: 1,
			conditions: [{ id: $.TRIGGERS.DIALOGUE.id, params: ["celebrate"] }],
			effects: [{ id: $.EFFECTS.SET_FLAG.id, params: [203] }, { id: $.EFFECTS.COSTUME.id, params: [1] }],
		},
		{
			category: "Erotic", subjectIds: [1, 3], targetId: 1,
			conditions: [{ id: $.TRIGGERS.KILL_ENEMY.id, params: [1] }],
			effects: [{ id: $.EFFECTS.SET_FLAG.id, params: [204] }],
			minLewdness: () => 5
		},
		{
			category: "Instant", subjectIds: [1], targetId: 1,
			conditions: [{ id: $.TRIGGERS.INSTANT.id, params: [] }],
			effects: [{ id: $.EFFECTS.CHANGE_VARIABLE.id, params: [602, 1] }],
			minLewdness: () => 6 + $gameVariables.value(602),
			requires: () => $.getClauseWithEffect({ id: $.EFFECTS.SET_FLAG.id, params: [204] }) && $gameVariables.value(602) < 3
		},
		{
			category: "Erotic", subjectIds: [1], targetId: 1,
			conditions: [{ id: $.TRIGGERS.DIALOGUE.id, params: ["masturbate"] }],
			effects: [{ id: $.EFFECTS.SET_FLAG.id, params: [205] }],
			requires: () => $.getClauseWithEffect({ id: $.EFFECTS.SET_FLAG.id, params: [204] })
		},
		{
			category: "Social", subjectIds: [3], targetId: 1,
			conditions: [{ id: $.TRIGGERS.DIALOGUE.id, params: ["gropeDate"] }],
			effects: [{ id: $.EFFECTS.SET_FLAG.id, params: [207] }],
			minLewdness: () => 5,
			requires: () => $.getClauseWithEffect({ id: $.EFFECTS.SET_FLAG.id, params: [203] })
		},
		// Clauses for Edgar to offer during quick negotiation
		{
			category: "Instant", subjectIds: [1], targetId: 3,
			conditions: [{ id: $.TRIGGERS.INSTANT.id, params: [] }],
			effects: [{ id: $.EFFECTS.TRADE_GOLD.id, params: [1, 500] }],
			requires: () => Liora.Edgar.getParty().gold() >= 1000 && Liora.Edgar.getParty().gold() < 1500
		},
		{
			category: "Combat", subjectIds: [3], targetId: 3,
			conditions: [{ id: $.TRIGGERS.RECEIVED.id, params: [12] }],
			effects: [{ id: $.EFFECTS.USE_SKILL.id, params: [37, 2] }, { id: $.EFFECTS.USE_SKILL.id, params: [1, 1] }],
		},
		{
			category: "Combat", subjectIds: [11], targetId: 3,
			conditions: [{ id: $.TRIGGERS.TURN.id, params: [2] }, { id: $.TRIGGERS.TURN.id, params: [3] }],
			effects: [{ id: $.EFFECTS.USE_SKILL.id, params: [45, 0] }],
		},
		{
			category: "Instant", subjectIds: [1], targetId: 3,
			conditions: [{ id: $.TRIGGERS.INSTANT.id, params: [] }],
			effects: [{ id: $.EFFECTS.TRADE_GOLD.id, params: [1, 1000] }],
			requires: () => Liora.Edgar.getParty().gold() >= 1500
		},
		{
			category: "Combat", subjectIds: [1, 2], targetId: 3,
			conditions: [{ id: $.TRIGGERS.RECEIVED.id, params: [12] }],
			effects: [{ id: $.EFFECTS.USE_SKILL.id, params: [45, 0] }],
		},
		{
			category: "Combat", subjectIds: [1], targetId: 3,
			conditions: [{ id: $.TRIGGERS.KILL_ENEMY.id, params: [0] }],
			effects: [{ id: $.EFFECTS.USE_SKILL.id, params: [1, 4] }],
		},
		{
			category: "Social", subjectIds: [1], targetId: 3,
			conditions: [{ id: $.TRIGGERS.DIALOGUE.id, params: ["minePyromantium"] }],
			effects: [{ id: $.EFFECTS.SET_FLAG.id, params: [206] }],
		},
		{
			category: "Combat", subjectIds: [2], targetId: 3,
			conditions: [{ id: $.TRIGGERS.KILL_ENEMY.id, params: [0] }],
			effects: [{ id: $.EFFECTS.USE_SKILL.id, params: [1, 3] }],
		},
	];

	// ---------------------------------------------------------------------------------
	// LOGIC
	// ---------------------------------------------------------------------------------

	// Checks if two entries in a condition / effect list are equal
	$.equalEntries = function(entry1, entry2) {
		if (entry1.id != entry2.id) return false;
		if (entry1.params.length != entry2.params.length) return false;
		for (let i = 0; i < entry1.params.length; ++i) {
			if (entry1.params[i] != entry2.params[i]) return false;
		}

		return true;
	}

	// Checks if a list of conditions / effects is a sublist of another
	$.subList = function(list1, list2) {
		return list2.every(entry2 => list1.some(entry1 => $.equalEntries(entry1, entry2)));
	}

	$.getClauseWithEffect = function(effect) {
		return $.gameClauses.find(clause => clause.hasEffect(effect));
	}

	$.getClauseWithEqualConditions = function(clause) {
		return $.gameClauses.find(other => clause.equalConditions(other.clause()));
	}

	$.existsSubsumingClause = function(clause) {
		const candidate = $.getClauseWithEqualConditions(clause);
		return candidate ? candidate.subsumesEffects(clause.clause()) : false;
	};

	$.removeSubsumedClauses = function(clause) {
		const toRemove = [];
		for (const other of $.gameClauses) {
			if (clause.subsumesConditions(other.clause())) {
				other.clause().effects = other.clause().effects.filter(effect => !clause.hasEffect(effect));
				if (other.effects().length == 0) {
					toRemove.push(other);
				}
			}
		}

		toRemove.forEach(other => $.gameClauses.remove(other));
	};

	$.existsStrictlyBetterClause = function(clause) {
		return $.gameClauses.some(other => {
			// If the clause has different conditions, then it's incomparable
			if (!other.subsumesConditions(clause)) return false;
			// If a contract clause exists that conflicts with this clause, then check if
			// adding this default clause would be a total increase in value
			const valueChange = other.mergeValue(clause);
			return valueChange <= 0;
		});
	};

	$.availableDefaultClauses = function() {
		return $.defaultClauses.filter(defaultClause =>
			(!defaultClause.requires || defaultClause.requires())
			&& !$.existsStrictlyBetterClause(defaultClause)
		).map(clause => new Game_Clause(clause).copy());
	};

	$.createEmptyClause = function(category, subjectId, targetId) {
		return new Game_Clause(
			{
				category: category,
				subjectId: subjectId,
				targetId: targetId,
				conditions: [],
				effects: []
			}
		);
	};

	$.createInitialContract = function() {
		return [new Game_Clause($.defaultClauses[0])];
	};

	$.addClause = function(clause) {
		const other = $.getClauseWithEqualConditions(clause);
		if (other) {
			other.merge(clause);
		} else {
			$.removeSubsumedClauses(clause);
			$.gameClauses.push(clause);
		}
	};

	class Game_Contract_Check {
		constructor(context) {
			this.context = context;
			this.results = {};
		}

		execute() {
			this.results.triggeredClauses = this.getTriggeredClauses();
			this.executeClauses(this.results.triggeredClauses);
			return this.hasTriggeredClause();
		}

		isTriggeredClause(clause) {
			return clause.clause().conditions.some(condition => this.context.conditionIds.includes(condition.id))
				&& clause.clause().conditions.every(condition => this.checkCondition(clause, condition));
		}

		hasTriggeredClause() {
			return this.results.triggeredClauses.length > 0;
		}

		getTriggeredClauses() {
			return $.gameClauses.filter(clause => this.isTriggeredClause(clause))
				.sort((clause1, clause2) => clause2.conditions().length - clause1.conditions().length);
		}

		checkCondition(clause, condition) {
			return $.dataTriggers[condition.id].checkCondition(this.context, clause, condition);
		}

		executeClauses(clauses) {
			const executedClauses = []
			for (const clause of clauses) {
				const conflicts = executedClauses.flatMap(executedClause => clause.conflictingEffects(executedClause));
				let executedAnyEffect = false;
				for (const effect of clause.effects()) {
					if (!conflicts.includes(effect)) {
						this.executeEffect(clause, effect);
						executedAnyEffect = true;
					}
				}
				executedClauses.push(clause.clause());

				if (!executedAnyEffect && this.results.triggeredClauses) {
					this.results.triggeredClauses.remove(clause);
				}
			}
		}

		executeEffect(clause, effect) {
			$.dataEffects[effect.id].execute(this.context, clause, effect);
			if (this.context.cannotExecute) {
				this.results.triggeredClauses = [];
			}
		}
	}

	class Game_Clause {
		constructor(clause) {
			this._clause = clause;
		}

		clause() {
			return this._clause;
		}

		category() {
			return this.clause().category;
		}

		subject() {
			return $gameActors.actor(this.clause().subjectId);
		}

		target() {
			return $gameActors.actor(this.clause().targetId);
		}

		conditions() {
			return this.clause().conditions;
		}

		trigger() {
			return this.clause().conditions[0];
		}

		restrictions() {
			return this.clause().conditions.slice(1) || [];
		}

		effects() {
			return this.clause().effects;
		}

		toString(contract = $.gameClauses, options = {}) {
			if (options.icons === undefined) options.icons = true;
			
			const headerText = this.toStringHeader(contract);
			const effectText = this.toStringEffects();
			const restrictionText = this.toStringRestrictions();
			if (options.restrictions === undefined) options.restrictions = restrictionText.length > 0;
			const clauseText = headerText + (options.restrictions ? [effectText, restrictionText].join("\n") : effectText);
			const clauseDetail = options.icons ? clauseText : clauseText.replace(/\\I\[[^\]]*\]/g, '');
			return clauseDetail.split("\\n").join("\n");
		}

		toStringHeader(contract) {
			const category = this.category();
			if (category != "Instant") {
				const subject = this.subject();
				const target = this.target();
				const index = $.categorizedClauses(contract)[category].indexOf(this);
				const categoryIndex = $.clauseCategories.indexOf(category);
				const paragraph = `\\c[10]§\\c[0]${categoryIndex + 1}.${index + 1}`;
				const triggerText = $.triggerClauseText.format(subject.name().toUpperCase(), this.toStringTrigger());
				const targetText = $.effectClauseText.format(target.name().toUpperCase());
				return paragraph + " " + triggerText + " " + targetText + "\n";
			}

			return "";
		}

		toStringTrigger() {
			const trigger = this.trigger();
			return trigger ? $.dataTriggers[trigger.id].toString(trigger) : $.PLACEHOLDER.toString();
		}

		toStringRestrictions() {
			const restrictionTexts = this.restrictions().map(restriction => `• ${$.dataTriggers[restriction.id].toString(restriction)}`);
			return restrictionTexts.join("\n");
		}

		toStringEffects() {
			const effectTexts = this.effects().map(effect => `• ${$.dataEffects[effect.id].toString(effect)}`)
			return effectTexts.length > 0 ? effectTexts.join("\n") : `•` + $.PLACEHOLDER.toString();
		}

		// Total negotiation value
		value() {
			const conditionValue = this.valueConditions();
			const effectValue = this.valueEffects();
			return effectValue > 0 ? Math.ceil((this.category() == "Instant" ? 1 : conditionValue) * effectValue) : "x" + conditionValue;
		}

		valueConditions() {
			return this.valueTrigger() * this.valueRestrictions();
		}

		valueTrigger() {
			const trigger = this.trigger();
			return trigger ? this.valueCondition(trigger) : 0;
		}

		valueRestrictions() {
			return this.restrictions().reduce((res, restriction) => res * this.valueCondition(restriction), 1);
		}

		valueCondition(condition) {
			return $.dataTriggers[condition.id].value(this, condition);
		}

		valueEffects() {
			return this.effects().reduce((res, effect) => res + this.valueEffect(effect), 0);
		}

		valueEffect(effect) {
			return $.dataEffects[effect.id].value(this, effect);
		}

		equalConditions(other) {
			if (this.conditions().length != other.conditions.length) return false;
			return this.subsumesConditions(other);
		}

		subsumesConditions(other) {
			if (this.clause().subjectId != other.subjectId) return false;
			if (this.clause().targetId != other.targetId) return false;
			return $.subList(other.conditions, this.conditions());
		}

		subsumesEffects(other) {
			return $.subList(this.effects(), other.effects);
		}

		hasEffect(effect) {
			return this.effects().some(other => $.equalEntries(effect, other));
		}

		hasCondition(condition) {
			return this.conditios().some(other => $.equalEntries(condition, other));
		}

		conflictingEffects(other) {
			const conflicts = [];
			const effects1 = this.effects();
			const effects2 = other.effects;
			for (const effect1 of effects1) {
				const conflict = effects2.find(effect2 => $.dataEffects[effect1.id].conflict(this, effect1, effect2));
				if (conflict) conflicts.push(effect1);
			}

			return conflicts;
		}

		conflictValue(other) {
			const conflicts = this.conflictingEffects(other);
			const conflictValue = conflicts.reduce((res, conflict) => res + this.valueEffect(conflict), 0);
			return conflictValue;
		}

		mergeValue(other) {
			const valueOther = other.effects.reduce((res, effect) => res + this.valueEffect(effect), 0);
			const conflictValue = this.conflictValue(other);
			return valueOther - conflictValue;
		}

		merge(other) {
			const conflicts = this.conflictingEffects(other.clause());
			conflicts.forEach(conflict => this.effects().remove(conflict));
			this.effects().push(...other.effects());
		}

		copy() {
			const data = {};
			Object.assign(data, this.clause());
			data.conditions = [...data.conditions];
			data.effects = [...data.effects];
			return new Game_Clause(data);
		}
	}

	$.Game_Clause = Game_Clause;

	$.categorizedClauses = function(contract = $.gameClauses) {
		return Liora.groupBy(contract, clause => clause.category());
	};

	$.clausesCategorizedByDay = function(clauses) {
		return Liora.groupBy(clauses, (clause) => clause.conditions()[1] ? clause.conditions()[1].params[0] : -1);
	}

	$.executeContract = function(context) {
		const contractCheck = new Game_Contract_Check(context);
		if (contractCheck.execute()) {
			$.queueContractMessage(contractCheck);
		}

		return contractCheck;
	};

	// Inject contract trigger condition checks
	$._Game_Action_apply = Game_Action.prototype.apply;
	Game_Action.prototype.apply = function(target) {
		$._Game_Action_apply.call(this, target);

		if (!target.isAlive()) {
			const subject = this.subject();
			if (subject.isActor() && target.isEnemy()) {
				const context = {
					subjectId: subject.actorId(),
					conditionIds: [$.TRIGGERS.KILL_ENEMY.id],
					params: [this, target]
				};
				$.executeContract(context);
			} else if (target.isActor()) {
				const context = {
					subjectId: target.actorId(),
					conditionIds: [$.TRIGGERS.DEFEATED.id],
					params: [this, subject]
				};
				$.executeContract(context);
			}
		}
	};

	// Registers a choice enablement condition in the message data
	PluginManager.registerCommand(PLUGIN_ID, "checkDialogueTriggers", args => {
		const subjectId = JSON.parse(args.subjectId);
		const tags = JSON.parse(args.tags);
		const context = {
			subjectId: subjectId,
			conditionIds: [$.TRIGGERS.DIALOGUE.id],
			tags: tags
		};
		$.executeContract(context);
	});

	// Inject trigger for turn start
	$._BattleManager_startInput = BattleManager.startInput;
	BattleManager.startInput = function() {
		const context = { conditionIds: [$.TRIGGERS.TURN.id] };
		$.executeContract(context);
		$._BattleManager_startInput.call(this);
	};

	// Inject trigger for receiving new state
	$._Game_BattlerBase_addNewState = Game_BattlerBase.prototype.addNewState
	Game_BattlerBase.prototype.addNewState = function(stateId) {
		$._Game_BattlerBase_addNewState.call(this, stateId);
		if (this.isActor()) {
			const context = {
				subjectId: this.actorId(),
				conditionIds: [$.TRIGGERS.RECEIVED.id],
				params: [stateId]
			};
			$.executeContract(context);
		}
	};

	// ---------------------------------------------------------------------------------
	// NEGOTIATION LOGIC
	// ---------------------------------------------------------------------------------

	$.beginNegotiate = function(category, actorId, proposals, valueFactor = 1) {
		$.negotiateClauses = [$.createEmptyClause(category, actorId, actorId)];
		if (proposals) {
			$.setupProposals(proposals);
			$.refreshProposalValues(valueFactor);
		}
	};

	$.setupProposals = function(proposals) {
		const startIndex = (3 * Liora.Calendar.currentWeekdayId()) % proposals.length;
		for (let i = 0; i < Math.min(3, proposals.length); ++i) {
			const index = (i + startIndex) % proposals.length;
			$.negotiateClauses[i + 1] = proposals[index];
		}
	};

	$.refreshProposalValues = function(valueFactor = 1) {
		for (let i = 1; i < $.negotiateClauses.length; ++i) {
			$gameVariables.setValue(45 + i, $.calcNegotiateValue(i) * valueFactor);
		}
	};

	$.endNegotiate = function(result) {
		if (result > 0) {
			const clauses = [$.negotiateClauses[0], $.negotiateClauses[result]];
			for (const clause of clauses) {
				if (clause.category() == "Instant") {
					const check = new Game_Contract_Check({});
					check.executeClauses([clause]);
				} else {
					$.addClause(clause);
				}
			}
		}
	};

	$.negotiateText = function(index) {
		return $.negotiateClauses[index].toString($.negotiateClauses).replace(/\\/g, "\x1B");
	};
	// Negotiate AI Constants
	$.calcNegotiateValue = function(clauseIndex) {
		const clause = $.negotiateClauses[clauseIndex];
		return clause.value();
	};

	$.availableProposals = function(actorId, filterMinLewdness = true) {
		return $.availableDefaultClauses().filter(clause =>
			clause.clause().targetId == actorId
			&& (!filterMinLewdness || actorId != 1 || ($gameVariables.value(5) >= clause.clause().minLewdness()))
		);
	};

	// Get available proposals which exceed the minimum value
	$.availableProposalsMin = function(actorId, valueFactor, minValue, filterMinLewdness) {
		return $.availableProposals(actorId, filterMinLewdness).filter(clause => clause.value() * valueFactor >= minValue);
	};

	// Get available proposals which cost at most the max value
	$.availableProposalsMax = function(actorId, valueFactor, maxValue) {
		return $.availableProposals(actorId).filter(clause => clause.value() * valueFactor <= maxValue);
	};

	$.proposalMinLewdness = function(index) {
		return $.negotiateClauses[index].clause().minLewdness();
	};

	$.isAcceptableProposal = function(index) {
		return $gameVariables.value(208) >= $gameVariables.value(207)
			&& $gameVariables.value(5) >= $.proposalMinLewdness(index)
			&& !$.existsSubsumingClause($.negotiateClauses[index]);
	};

	$.setNegotiateTrigger = function(triggerId, params, newSubjectId) {
		$.negotiateClauses[0].clause().conditions[0] = { id: triggerId, params: params };
		if (newSubjectId) {
			$.negotiateClauses[0].clause().subjectId = newSubjectId;
		}
	};

	$.setNegotiateEffect = function(effectId, params) {
		$.negotiateClauses[0].clause().effects[0] = { id: effectId, params: params };
	};

	// ---------------------------------------------------------------------------------
	// PERSISTANCE
	// ---------------------------------------------------------------------------------

	// Initialize the contract data
	$._DataManager_setupNewGame = DataManager.setupNewGame;
	DataManager.setupNewGame = function() {
		$._DataManager_setupNewGame.call(this);
		$.gameClauses = $.createInitialContract();
	};

	// Save the contract data
	$._DataManager_makeSaveContents = DataManager.makeSaveContents;
	DataManager.makeSaveContents = function() {
		const contents = $._DataManager_makeSaveContents();
		contents.contracts = { "liora": $.gameClauses.map(clause => clause.clause()) }
		if ($.negotiateClauses) {
			contents.contracts.negotiateContract = $.negotiateClauses.map(clause => clause.clause());
		}
		return contents;
	};

	// Load the contract data
	$._DataManager_extractSaveContents = DataManager.extractSaveContents;
	DataManager.extractSaveContents = (contents) => {
		$._DataManager_extractSaveContents(contents);
		$.gameClauses = contents.contracts?.["liora"].map(clause => new Game_Clause(clause));
		if (contents.contracts?.negotiateContract) {
			$.negotiateClauses = contents.contracts.negotiateContract.map(clause => new Game_Clause(clause));
		}
	};

	// ---------------------------------------------------------------------------------
	// COMPENDIUM
	// ---------------------------------------------------------------------------------

	// Inject Contracts Category into Compendium
	$._Window_CompendiumCategory_makeCommandList = Window_CompendiumCategory.prototype.makeCommandList;
	Window_CompendiumCategory.prototype.makeCommandList = function() {
		this.addCommand($.contractsCommand, $.compendiumCategory);
		$._Window_CompendiumCategory_makeCommandList.call(this);
	};

	// Inject Contract Icon into Compendium
	$._Window_CompendiumCategory_getIcons = Window_CompendiumCategory.prototype.getIcons;
	Window_CompendiumCategory.prototype.getIcons = function() {
		return [190].concat($._Window_CompendiumCategory_getIcons.call(this));
	}

	class Window_ContractList extends Window_Compendium {
		constructor(rect) {
			super(rect);
			this._expandedCategories = [];
			this.makeClauses();
		}

		isExpanded(category) {
			return this._expandedCategories[category];
		}

		flipExpansionState(category) {
			this._expandedCategories[category] = !this._expandedCategories[category];
			this.makeClauses();
		}

		makeClauses() {
			this._data = this.createClauseList();
		}

		createClauseList() {
			const array = [];
			const categorizedClauses = $.categorizedClauses();
			for (const category of $.clauseCategories) {
				array.push(category);
				const clauses = categorizedClauses[category];
				if (clauses) {
					const dayGroupedClauses = $.clausesCategorizedByDay(clauses);
					delete dayGroupedClauses[-1];
					if (this.isExpanded(category)) {
						Object.keys(dayGroupedClauses).forEach(weekdayId => array.push({ category: category, weekdayId: weekdayId }));
					}
				}
			}
			return array;
		}

		update() {
			this.makeClauses();
			super.update();

			if (this.visible && this._mapPageToWindow[0]) {
				this._mapPageToWindow[0].show();
				const item = this.item();
				if (item) {
					this._mapPageToWindow[0].updateContract(item);
				}
			}
		}

		drawItem(index) {
			const item = this.itemAt(index);
			const rect = this.itemLineRect(index);
			rect.y -= 9;
			this.contents.fontFace = 'franklin-gothic-med';
			if (item.weekdayId) {
				// This is a weekday restriction
				this.contents.fontSize = 20;
				const weekday = Liora.Calendar.weekday(item.weekdayId);
				this.drawIcon(Liora.Calendar.weekdayIconIndex(weekday), rect.x + 15, rect.y + 5);
				this.drawGradientText(`${weekday.name_cte.toUpperCase()}`, this._standardGradient, rect.x + 49, rect.y, rect.width, "left", this._textOptions);
			} else {
				// This is a category
				const hasRestrictedClauses = $.categorizedClauses()[item]?.some(clause => clause.conditions()[1]);
				const sign = hasRestrictedClauses ? (this.isExpanded(item) ? "-" : "+") : "";
				this.contents.fontSize = 24;
				this.drawGradientText(sign, this._standardGradient, rect.x, rect.y, rect.width, "left", this._textOptions);
				let offsetX = this.textWidth("+");
				this.drawGradientText(`§${index + 1}`, this._standardGradient, rect.x + offsetX, rect.y, rect.width, "left", this._textOptions);
				offsetX += this.textWidth(`§${index + 1} `);
				const categoryId = $.clauseCategories.indexOf(item);
				const category = TextManager.message(`${PLUGIN_ID}-clauseCategories_cte-${categoryId}`) || item;
				this.contents.fontSize = 22;
				this.drawGradientText(category, this._standardGradient, rect.x + offsetX, rect.y, rect.width, "left", this._textOptions);
			}
		}

		addPageWindow(window) {
			this._mapPageToWindow.push(window);
		}
	}

	class Window_ContractDetail extends Window_CompendiumDetail {
		updateContract(item) {
			this.contents.clear();
			this.makeClauseList(item);
			this.refresh();
		}

		makeClauseList(item) {
			if (item.weekdayId) {
				const clauses = $.categorizedClauses()[item.category];
				this._data = $.clausesCategorizedByDay(clauses)[item.weekdayId];
			} else {
				// Only take unrestricted clauses in the top level categories
				const clauses = $.categorizedClauses()[item];
				this._data = clauses ? $.clausesCategorizedByDay(clauses)[-1] : [];
			}
		}

		drawItem(index) {
			const item = this._data[index];
			const rect = this.itemRect(index);
			const clauseText = item.toString($.gameClauses, { restrictions: false});
			this.contents.fontFace = "franklin-gothic-med-cond";
			this.contents.fontSize = 23;
			this.drawTextEx(clauseText, rect.x + 10, rect.y, rect.width, true);
		}

		drawItemBackground(index) {
			Window_Command.prototype.drawItemBackground.call(this, index);
		}
	}

	// ---------------------------------------------------------------------------------
	// POPUP WINDOW
	// ---------------------------------------------------------------------------------

	$.queueContractMessage = function(contractCheck) {
		SceneManager._scene.queueContractMessageWindow(contractCheck);
	};

	// Window that pops up upon changing the state of a quest
	class Window_ContractMessage extends Window_Base {
		constructor(parent, rect) {
			super(rect);
			this.opacity = 255;
			this.contentsOpacity = 255;
			this._showCount = 0;
			this._message = "";
			this._parent = parent;
			this.setBackgroundType(1);
			this.hide();
		}

		lineHeight = function() {
			return 26;
		};

		update() {
			super.update();
			if (this._showCount > 0) {
				if (this.x < -this._parent._windowLayer.x) {
					this.x += 6;
				} else {
					this.x = -this._parent._windowLayer.x;
				}
				this.updateFadeIn();
				this._showCount--;
			} else if (this.visible) {
				if (this.x > this._parent._windowLayer.x - this.width * 2) {
					this.x -= 1;
				}
				this.updateFadeOut();
			}
		}

		updateFadeIn() {
			this.opacity += 16;
			this.contentsOpacity += 16
		}

		updateFadeOut() {
			this.opacity -= 4;
			this.contentsOpacity -= 5;

			if (this.opacity <= 0) {
				this._parent._windowLayer.removeChild(this);
				this._parent._contractMessageWindows.remove(this);
				this.destroy();
			}
		}

		open(message) {
			this.show();
			this._showCount = 180;
			this._message = message;
			this.refresh();
		}

		close() {
			this._showCount = 0;
			this._parent._windowLayer.removeChild(this);
			this._parent._contractMessageWindows.remove(this);
			this.destroy();
		}

		refresh() {
			this.contents.clear();
			const width = this.textWidth(this._message);
			this.contents.fontFace = "franklin-gothic-med-cond";
			this.contents.fontSize = 18;
			this.drawTextEx(this._message, 0, 0, width, "left", true);
		}
	}

	$._Scene_Compendium_createPages = Scene_Compendium.prototype.createPages;
	Scene_Compendium.prototype.createPages = function() {
		this._contractWindow = this.createPage(Window_ContractList, this.onContractPageOk);
		this._contractDetailWindow = this.createDetailPage(Window_ContractDetail, this._contractWindow);
		$._Scene_Compendium_createPages.call(this);
	}

	Scene_Compendium.prototype.onContractPageOk = function() {
		this._contractWindow.activate();

		const item = this._contractWindow.item();
		if (!item.name_cte) {
			this._contractWindow.flipExpansionState(item);
			this._contractWindow.refresh();
		}

		this._bestiaryWindow.activate();
	}

	$._Scene_Message_createAllWindows = Scene_Message.prototype.createAllWindows;
	Scene_Message.prototype.createAllWindows = function() {
		$._Scene_Message_createAllWindows.call(this);
		this._contractMessageWindows = [];
	};

	// Create a popup window for displaying triggered clauses
	Scene_Message.prototype.queueContractMessageWindow = function(contractCheck) {
		SoundManager.playUseItem();
		for (const triggeredClause of contractCheck.results.triggeredClauses) {
			const clauseText = triggeredClause.toString($.gameClauses, { icons: false});
			const rect = this.contractMessageWindowRect(clauseText.split("\n").length);
			const contractMessageWindow = new Window_ContractMessage(this, rect);
			contractMessageWindow.open(clauseText);
			this.addWindow(contractMessageWindow);
			this._contractMessageWindows.push(contractMessageWindow);
		}
	};

	// Creates the size dimension sof the quest popup window
	Scene_Message.prototype.contractMessageWindowRect = function(lines) {
		const ww = 385;
		const wh = lines == 2 ? 65 : 85;
		const wx = this._windowLayer.x - ww * 2;
		const wy = this._contractMessageWindows.reduce((res, window) => res + window.height, 130);
		return new Rectangle(wx, wy, ww, wh);
	};
})(Liora.Contract);

