//=============================================================================
// RPG Maker MZ - Liora Edgar
// ----------------------------------------------------------------------------
// (C)2025 aura-dev
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
// ----------------------------------------------------------------------------
// Version
// 1.0.0 02/04/2025
//=============================================================================

/*:
 * @target MZ
 * @plugindesc Liora Edgar
 * @author aura-dev
 * @url https://gitgud.io/auragamedev/auramz
 *
 * @help liora_edga.js
 *
 * Contains game customizations for Edgar specific game rules.
 *
 * @param claimItemText_cte
 * @text Claim Item Text
 * @desc Text displayed when an obtained item was claimed by someone.
 * @default \c[10](Claimed by %1)\c[0]
 * 
 * @param prestigeLabel_cte
 * @text Prestige Label
 * @desc Text displayed in the Status Menu for the Prestige stat
 * @default Prestige
 */

window.Liora = window.Liora || {};
Liora.Edgar = {};

(($) => {
	const PLUGIN_ID = "liora_edgar";
	$.Params = PluginManager.parameters(PLUGIN_ID);

	Object.defineProperties($, {
		"claimItemText": { get: () => TextManager.message(`${PLUGIN_ID}-claimItemText_cte`) || $.Params.claimItemText_cte },
		"prestigeLabel": { get: () => TextManager.message(`${PLUGIN_ID}-prestigeLabel_cte`) || $.Params.prestigeLabel_cte }
	});

	$.actorId = 3;
	$.prestige = function() {
		return $gameVariables.value(214);
	}
	$.prestigeDecay = function() {
		return $gameVariables.value(215);
	}
	$.prestigeDecayFactor = function() {
		return 1 + $gameVariables.value(215);
	}

	$.getActor = function() {
		return $gameActors.actor($.actorId);
	};

	// Battle AI Constants
	$.killFactor = 5;
	$.bossKillFactor = $.killFactor * 10;
	$.randomness = 2;
	$.hpValueFactor = 2;

	// Give Edgar his own party so he can have his own gold and inventory
	$.getParty = function() {
		const actor = $.getActor();
		if (!actor.soloParty) {
			actor.soloParty = new Game_Party();
		}

		return actor.soloParty;
	}

	// Inject addition of guard action;
	$._Game_Actor_makeActionList = Game_Actor.prototype.makeActionList;
	Game_Actor.prototype.makeActionList = function() {
		const list = $._Game_Actor_makeActionList.call(this);
		const guardAction = new Game_Action(this);
		guardAction.setSkill(2);
		list.push(guardAction);
		return list;
	};

	// Removes normalization and instead returns value and max value to
	// normalize the total value later
	Game_Action.prototype.evaluateWithTarget = function(target) {
		if (this.isHpEffect()) {
			const value = this.makeDamageValue(target, this.subject().isStateAffected(12));
			if (this.isForOpponent()) {
				return [Math.min(value, target.hp), Math.max(target.hp, 1)];
			} else {
				const recovery = Math.min(-value, target.mhp - target.hp);
				return [recovery, target.mhp];
			}
		}

		return [0, 0];
	};

	$._Game_Actor_makeAutoBattleActions = Game_Actor.prototype.makeAutoBattleActions
	Game_Actor.prototype.makeAutoBattleActions = function() {
		if ($._forceSkill) {
			const action = new Game_Action(this);
			action.setSkill($._forceSkill[0]);
			let targets = action.itemTargetCandidates();
			if (targets.length > 0 || $._forceSkill[1] == 0) {
				let validTarget = false;
				switch ($._forceSkill[1]) {
					case 0: // Self
						this._targetIndex = 0;
						validTarget = true;
						break;
					case 2: // Random Boss
						targets = targets.filter(target => target.enemy().meta.boss);
					case 1: // Random Enemy
						if (targets.length == 0) break;
						this._targetIndex = targets[Math.randomInt(targets.length)].index();
						validTarget = true;
						break;
					case 3: // Liora's last target
						validTarget = action.setTargetLastTargetOfActor(1);
						break;
					case 4: // Rean's last target
						validTarget = action.setTargetLastTargetOfActor(2);
						break;
				}

				if (validTarget) {
					this.setAction(0, action);
					this.setActionState("waiting");
					return;
				}
			}
		}

		$._Game_Actor_makeAutoBattleActions.call(this);
	};

	// Tries to set the target of this action as the last target of the actor
	// return false if its not a valid target
	Game_Action.prototype.setTargetLastTargetOfActor = function(actorId) {
		const actor = $gameActors.actor(actorId);
		if (actor._lastTargetIsEnemy && actor._lastTargetIsEnemy == this.isForOpponent() && $gameTroop.members()[actor._lastTargetIndex].isAlive()) {
			this._targetIndex = actor._lastTargetIndex;
			return true;
		}

		return false;
	};

	// Custom AI behavior for Edgar
	Game_Action.prototype.evaluate = function() {
		const item = this.item();
		let value = 0;
		switch (item.id) {
			case 2:
				if ($.prestigeDecay() > 0) {
					// Reduce guard priority proportionally to prestige decay
					value = $gameParty.aliveMembers().length > 1 ? (1 / $.prestigeDecayFactor()) : -1;
				} else {
					// By default, make Guard the highest priority action,
					// except when the only one alive
					value = $gameParty.aliveMembers().length > 1 ? 2 : -1;
				}
				break;
			case 4:
				if ($.prestigeDecay() > 0) {
					// Reduced chance of using a retreat 
					value = $gameParty.aliveMembers().length > 1 ? -1 : (1 / $.prestigeDecayFactor());
				} else {
					// Give a chance to retreat if alone and cannot kill
					value = $gameParty.aliveMembers().length > 1 ? -1 : 2;
				}
				break;
			default:
				let maxValue = 0;
				const costs = this.subject().actorSkillCost(item);
				const costValue = costs ? -costs.value * (costs.currency == "hp" ? ($.hpValueFactor / this.subject().hp) : 1 / this.subject().mp) : 0;
				if (this.isForAll()) {
					value -= costValue;
				}
				for (const target of this.itemTargetCandidates()) {
					const targetKillFactor = target.isEnemy() && target.enemy().meta.boss ? $.bossKillFactor : $.killFactor;
					let [targetValue, targetMaxValue] = this.evaluateWithTarget(target);
					if (targetValue>= targetMaxValue) {
						// Drastically increase value if the attack would kill a target
						targetValue *= targetKillFactor;
					} else if ($.prestigeDecay() > 0 && target.isEnemy() && target.enemy().meta.area_boss) {
						// When fighting an area boss while prestige decay is active multiply by prestige decay
						targetValue *= $.prestigeDecayFactor();
					}

					if (this.isForAll()) {
						value += targetValue;
						maxValue += targetMaxValue;
					} else if (targetValue > value) {
						value = targetValue + costValue;
						maxValue += targetMaxValue + costValue;
						this._targetIndex = target.index();
					}
				}
				// Normalize to 1
				value /= maxValue;
		}

		value *= this.numRepeats();
		if (value > 0) {
			value += Math.random() * (1 - 1 / $.randomness);
		}
		return value;
	};

	// Check for claimed items
	$._Game_Enemy_makeDropItems = Game_Enemy.prototype.makeDropItems;
	Game_Enemy.prototype.makeDropItems = function() {
		const dropItems = $._Game_Enemy_makeDropItems.call(this);
		// If Edgar has a claim on the drop items
		if (this._actorClaimDropItems && this._actorClaimDropItems == $.actorId) {
			BattleManager._edgarRewards.push(...dropItems);
		}

		return dropItems;
	}

	// Initialize data structure for tracking edgar drops
	$._BattleManager_makeRewards = BattleManager.makeRewards;
	BattleManager.makeRewards = function() {
		this._edgarRewards = [];
		$._BattleManager_makeRewards.call(this);
	}

	// Inject logic to display item claims
	$._BattleManager_displayDropItems = BattleManager.prototype.displayDropItems;
	BattleManager.displayDropItems = function() {
		const items = this._rewards.items;
		if (items.length > 0) {
			$gameMessage.newPage();
			const groupedItems = Liora.groupBy(items, item => item.id);
			const edgarItems = Liora.groupBy(this._edgarRewards, item => item.id);
			for (let [ itemId, itemGroup ] of Object.entries(groupedItems)) {
				const item = $dataItems[itemId];
				const edgarClaims = edgarItems[itemId]?.length || 0;
				const count = itemGroup.length - edgarClaims;
				if (count > 0) {
					const text = TextManager.obtainItem.format(item.name, count);
					$gameMessage.add(text);
				}
				if (edgarClaims > 0) {
					const actor = $gameActors.actor($.actorId);
					const text = TextManager.obtainItem.format(item.name, edgarClaims);
					$gameMessage.add(text + " " + $.claimItemText.format(actor.name()));
				}
			}
		}
	};

	// Inject logic to split up dropped items based on claims
	$._BattleManager_gainDropItems = BattleManager.gainDropItems;
	BattleManager.gainDropItems = function() {
		for (const item of this._edgarRewards) {
			const index = this._rewards.items.indexOf(item);
			if (index != -1) {
				this._rewards.items.splice(index, 1);
			}
			$.getParty().gainItem(item, 1);
		}
		$._BattleManager_gainDropItems.call(this);
	};

	// Cleanup force skill variable
	$._BattleManager_startInput = BattleManager.startInput;
	BattleManager.startInput = function() {
		$._BattleManager_startInput.call(this);
		$._forceSkill = undefined;
	};

	// Cleanup force skill variable
	$._BattleManager_endBattle = BattleManager.endBattle;
	BattleManager.endBattle = function(result) {
		$._BattleManager_endBattle.call(this, result);
		$._forceSkill = undefined;
	};

	// Memorize the type of the last target
	$._Game_Battler_setLastTarget = Game_Battler.prototype.setLastTarget;
	Game_Battler.prototype.setLastTarget = function(target) {
		$._Game_Battler_setLastTarget.call(this, target);
		this._lastTargetIsEnemy = target.isEnemy();
	};
})(Liora.Edgar);

