import json
import os
from utils.config import LANGUAGES_FILE, BASE_DIR, PUBLIC_DIR, DEFAULT_LANGUAGE

"""Counts translated vs untranslated keys in JSON files within the given folder."""
def count_translations(folder):
    def count_keys(data, countEmpty):
        if isinstance(data, dict):
            return sum(count_keys(v, countEmpty) for v in data.values())
        if countEmpty is True or data != "":
            return 1
        else:
            return 0
            
    
    translation_data = {}
    for root, _, files in os.walk(folder):
        for file in sorted(files):
            if file.endswith(".json"):
                file_path = os.path.join(root, file)
                with open(file_path, "r", encoding="utf-8") as f:
                    try:
                        data = json.load(f)
                        total_keys = count_keys(data, True)
                        translated_keys = count_keys(data, False)
                        fileKey = file_path[len(folder)+1:]
                        translation_data[fileKey] = {
                            "translated": translated_keys,
                            "total": total_keys
                        }
                    except json.JSONDecodeError:
                        print(f"Warning: Could not parse {file_path}")
    return translation_data

"""Generates an HTML dashboard displaying translation progress."""
def generate_dashboard(languages):
    print("Gathering data...")
    dashboard_data = {}
    for lang in languages:
        if (lang != DEFAULT_LANGUAGE):
            lang_folder = os.path.join(BASE_DIR, lang)
            print(f'...Gathering data for langauge {lang_folder}...')
            if os.path.exists(lang_folder):
                dashboard_data[lang] = count_translations(lang_folder)
        html = """
<html>
<head>
    <title>Translation Progress</title>
    <style>
        body { font-family: Arial, sans-serif; }
        .folder { cursor: pointer; font-weight: bold; }
        .hidden { display: none; }
        .progress-bar-container { width: 100%; background-color: #ddd; border-radius: 5px; margin: 5px 0; }
        .progress-bar { height: 20px; background-color: green; border-radius: 5px; text-align: center; color: white; font-weight: bold; }
        .search-box { margin-bottom: 20px; }
    </style>
    <script>
        function toggleVisibility(id, toggleId) {
            const element = document.getElementById(id);
            const toggle = document.getElementById(toggleId);
            if (element.classList.contains("hidden")) {
                element.classList.remove("hidden");
                toggle.innerHTML = "[-]";
            } else {
                element.classList.add("hidden");
                toggle.innerHTML = "[+]";
            }
        }
        
        function filterFiles() {
            const searchValue = document.getElementById("search").value.toLowerCase();
            const hideTranslated = document.getElementById("hideTranslated").checked;
            const hideUntranslated = document.getElementById("hideUntranslated").checked;
            const rows = document.querySelectorAll("tr.file-row");

            rows.forEach(row => {
                const fileName = row.getAttribute("data-filename").toLowerCase();
                const progress = parseFloat(row.getAttribute("data-progress"));

                const matchesSearch = searchValue === "" || fileName.includes(searchValue);
                const matchesFilter = (!hideTranslated || progress < 100) && (!hideUntranslated || progress >= 100);

                if (matchesSearch && matchesFilter) {
                    row.style.display = "";
                } else {
                    row.style.display = "none";
                }
            });
        }
    </script>
</head>
<body>
    <h1>Translation Progress</h1>
    <div class="search-box">
        <input type="text" id="search" placeholder="Search for a file..." onkeyup="filterFiles()">
        <label>
            <input type="checkbox" id="hideTranslated" onchange="filterFiles()"> Hide fully translated files
        </label>
        <label>
            <input type="checkbox" id="hideUntranslated" onchange="filterFiles()"> Hide incompletely translated files
        </label>
    </div>
"""
        
        for lang, files in dashboard_data.items():
            total_translated = sum(f["translated"] for f in files.values())
            total_keys = sum(f["total"] for f in files.values())
            if (total_keys > 0):
                progress = total_translated / total_keys * 100
                html += f'<div class="folder" onclick="toggleVisibility(\'{lang}\', \'toggle_{lang}\')">'
                html += f'<span id="toggle_{lang}">[+]</span> {lang} - {progress:.2f}%</div>'
                html += f'<div class="progress-bar-container"><div class="progress-bar" style="width: {progress:.2f}%">{progress:.2f}%</div></div>\n'
                html += f'<div id="{lang}" class="hidden">\n'
                
                # File-specific translation stats
                html += '<table style="margin-left: 20px; border-collapse: collapse;">'
                html += '''
<tr>
    <th style="text-align: left;">File</th>
    <th style="text-align: right;">Translated</th>
    <th style="text-align: right;">Total</th>
    <th style="text-align: right;">Progress</th>
</tr>
'''
                
                for file, stats in files.items():
                    if (stats["total"] > 0):
                        file_progress = stats["translated"] / stats["total"] * 100
                        html += f'''
<tr class="file-row" data-filename="{file}" data-progress="{file_progress}">
    <td>{file}</td>
    <td style="text-align: right;">{stats["translated"]}</td>
    <td style="text-align: right;">{stats["total"]}</td>
    <td style="text-align: right;">{file_progress:.2f}%</td>
</tr>
'''
                html += '</table></div>\n'
    html += """</body></html>"""
    if not os.path.exists(PUBLIC_DIR):
        os.makedirs(PUBLIC_DIR)
    dashboard_path = os.path.join(PUBLIC_DIR, "index.html")
    with open(dashboard_path, "w", encoding="utf-8") as f:
        print("Printing HTML...")
        f.write(html)
    print("...Done!")
            
def main():
    with open(LANGUAGES_FILE, "r", encoding="utf-8") as f:
        languages = json.load(f)
    generate_dashboard(languages)

if __name__ == "__main__":
    main()